/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "traverse.h"
#include "logw.h"
#include "replicaimpl.h"
#include "cfsrw.h"
#include "cfsw_record.h"
#include "cfsw_prot.h"
#include "cfsr.h"
#include "cfsr_prot.h"

class cfsrw_impl : public cfsrw {
private:
  bigint _projid;
  bigint _head;
  bigint _replicaaddr;
  ptr<sfspriv> _sk;
  ptr<logw> _logw;
  ptr<projinterface> _projinf;
  ptr<replica_impl> _replica;
  bigint _verf;
  comvec _commits;
  bool _anonymous;

  void load_proj (cbcfsrw, ptr<head_state> proj, log_status);
  void fetch_proj (cbcfsrw cb, ptr<head_state> v);
  void load_replica (cbcfsrw cb, ptr<head_state> st, bool, cfsrw_status status);

  //
  // CFSW
  //

  void reply_now (cfswcallp c, cfsrw_status status);
  void reply (cfswcallp c, cfsrw_status status);
  void update_replica_pre (cfswcallp c, cfsrw_status status);
  void update_replica_post (cfswcallp c, cfsrw_status status);
  void insert_replica_load (cbcfsrw cb, bigint addr, cfsrw_status status);
  void insert_replica_pre (cfswcallp c, bigint addr, cfsrw_status status);
  void insert_replica_post (cfswcallp c, bigint addr, cfsrw_status status);
  void update_pre_finished (cfswcallp c, cfsrw_status status);

  void lock_cb (cfswcallp c, comvec v, vervec pkv, log_status status);
  void anon_lock_cb (cfswcallp c, ptr<head_state> v, log_status s);
  void abort_cb (cfswcallp c, cfsrw_status s, log_status status);
  void commit_cb (cfswcallp c, bigint, log_status status);

public:
  cfsrw_impl (bigint va, bigint loga, ref<sfspriv> sk, cbcfsrw cb);
  cfsrw_impl (bigint va, cbcfsrw cb);
  virtual ~cfsrw_impl () {}

  ptr<projinterface> getprojinf () const { return _projinf; }
  ptr<replica> getreplica () const { return _replica; }

  void dispatch (cfswcallp c);
  void dispatch (cfsrcallp c);
};

ref<cfsrw>
cfsrw::alloc (bigint va, cbcfsrw cb)
{
  return New refcounted<cfsrw_impl> (va, cb);
}

ref<cfsrw>
cfsrw::alloc (bigint va, bigint loga, ref<sfspriv> sk, cbcfsrw cb)
{
  return New refcounted<cfsrw_impl> (va, loga, sk, cb);
}

cfsrw_impl::cfsrw_impl (bigint va, bigint loga, ref<sfspriv> sk, cbcfsrw cb)
  : _projid (va), _head (loga), _sk (sk), _anonymous (false)
{
  _replicaaddr = _projid;
  char xxb[20];
  rnd.getbytes (xxb, 20);
  mpz_set_rawmag_be (&_verf, xxb, 20);
  _projinf = New refcounted<projinterface>(va);
  _projinf->fetch_heads(wrap (this, &cfsrw_impl::load_proj, cb));
}

cfsrw_impl::cfsrw_impl (bigint va, cbcfsrw cb)
  : _projid (va), _anonymous (true)
{
  _replicaaddr = _projid;
  char xxb[20];
  rnd.getbytes (xxb, 20);
  mpz_set_rawmag_be (&_verf, xxb, 20);
  _projinf = New refcounted<projinterface>(va);
  _projinf->fetch_heads(wrap (this, &cfsrw_impl::load_proj, cb));
}

void
cfsrw_impl::load_proj (cbcfsrw cb, ptr<head_state> st, log_status status)
{
  if (status)
    cb (CFSRWERR_NOPROJECT);
  else {
    if (!_anonymous)
      _logw = logw::alloc (_head, _sk, _verf, _projinf, _projid);
    cfsr::init ();
    _replica = New refcounted<replica_impl> (_projid, _replicaaddr);
    _replica->load (wrap (this, &cfsrw_impl::load_replica, cb, st, false));
  }
}

void
cfsrw_impl::load_replica (cbcfsrw cb, ptr<head_state> st, bool build, cfsrw_status status)
{
  if (status && !build) {
    warn << "no replica, building from scratch\n";
    assert (_projinf);
    //_projblock->fetch (wrap (this, &cfsrw_impl::fetch_projblock, cb));
    fetch_proj(cb, st);
  }
  else {
    if (status == CFSRW_OK && build) {
      _replica->sync (wrap (this, &cfsrw_impl::insert_replica_load, cb));
    } else
      cb (status);
  }
}

void
cfsrw_impl::fetch_proj (cbcfsrw cb, ptr<head_state> st)
{
  _replica->update
    (st->commits, wrap (this, &cfsrw_impl::load_replica, cb, st, true));
}


//
// CFSW
//

#define SYMLINK_MAX_LENGTH 1024
#define CONCURRENT_OPS 8

void
cfsrw_impl::reply_now (cfswcallp c, cfsrw_status status)
{
  c->reply (status);
}

void
cfsrw_impl::reply (cfswcallp c, cfsrw_status status)
{
  delaycb (0, wrap (this, &cfsrw_impl::reply_now, c, status));
}

void
cfsrw_impl::update_pre_finished (cfswcallp c, cfsrw_status status)
{
  c->reply (status);
}

void
cfsrw_impl::insert_replica_load (cbcfsrw cb, bigint addr, cfsrw_status status)
{
  cb (status);
}

void
cfsrw_impl::insert_replica_pre (cfswcallp c, bigint addr, cfsrw_status status)
{
  if (status) {
    update_pre_finished (c, CFSRWERR_NOREPLICA);
  } else
    update_pre_finished (c, CFSRW_OK);
}

void
cfsrw_impl::insert_replica_post (cfswcallp c, bigint addr, cfsrw_status status)
{
  if (status){
    reply (c, CFSRWERR_NOREPLICA);
  } else
    reply (c, CFSRW_OK);
}

void
cfsrw_impl::update_replica_pre (cfswcallp c, cfsrw_status status)
{
  if (!status) {
    if (_replica->changed ()) {
      _replica->sync (wrap (this, &cfsrw_impl::insert_replica_pre, c));
    } else
      update_pre_finished (c, CFSRW_OK);
  }
  else {
    warn << "cannot build snapshot after acquiring lock\n";
    if (c->proc () == CFSWPROC_WLOCK) {
      _logw->abort
	(wrap (this, &cfsrw_impl::abort_cb, c, CFSRWERR_NOREPLICA));
    } else {
      reply (c, CFSRWERR_NOREPLICA);
    }
  }
}

void
cfsrw_impl::update_replica_post (cfswcallp c, cfsrw_status status)
{
  if (!status) {
    if (_replica->changed ()) {
      _replica->sync (wrap (this, &cfsrw_impl::insert_replica_post, c));
    } else
      reply (c, CFSRW_OK);
  }
  else {
    reply (c, CFSRWERR_NOREPLICA);
  }
}

void
cfsrw_impl::lock_cb (cfswcallp c, comvec v, vervec pkv, log_status status)
{
  if (status == LOG_OK) {
    _commits = v;
    _replica->setpkv (pkv);
    _replica->update (v, wrap (this, &cfsrw_impl::update_replica_pre, c));
  }

  else if (status == LOGERR_EXCLCONT)
    reply (c, CFSRWERR_EXCLCONT);
  else
    reply (c, CFSRWERR_IO);
}

void
cfsrw_impl::abort_cb (cfswcallp c, cfsrw_status s, log_status status)
{
  if (status != LOGERR_ABORTED)
    reply (c, CFSRWERR_IO);
  else
    reply (c, s);
}

void
cfsrw_impl::commit_cb (cfswcallp c, bigint n, log_status status)
{
  if (status)
    reply (c, CFSRWERR_IO);
  else {
    _commits.add (_head, n);
    _replica->update
      (_commits, wrap (this, &cfsrw_impl::update_replica_post, c));    
  }
}

void
cfsrw_impl::anon_lock_cb (cfswcallp c, ptr<head_state> v, log_status s)
{
  if (s)
    reply (c, CFSRWERR_IO);
  else {
    assert (v);
    _commits = v->commits;
    vervec pkv;
    for (unsigned i=0; i < v->lptrs.size (); i++)
      pkv.add (v->pkv.head (i), v->pkv.seqn (i));
    _replica->setpkv (pkv);
    _replica->update
      (_commits, wrap (this, &cfsrw_impl::update_replica_pre, c));
  }
}

void
cfsrw_impl::dispatch (cfswcallp c)
{
  cfsw_proc proc = c->proc ();

  if (_anonymous && proc != CFSWPROC_RLOCK) {
    reply (c, CFSRWERR_NOLOG);
    return;
  }

  switch (proc) {
    case CFSWPROC_NULL:
      reply (c, CFSRW_OK);
      break;

    case CFSWPROC_SETATTR:
      {
	cfsw_setattr_arg *a = static_cast<cfsw_setattr_arg *> (c->arg ());
	deltap new_record =
	  New refcounted<attr_delta> (a->object, a->new_attributes);
	_logw->append (new_record);
	reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_UNLINK:
      {
	cfsw_unlink_arg *a = static_cast<cfsw_unlink_arg *> (c->arg ());
        deltap new_record = New refcounted<unlink_delta>
	  (a->object, a->type, a->dir, a->name);
        _logw->append (new_record);
	reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_PATCH:
      {
	cfsw_patch_arg *a = static_cast<cfsw_patch_arg *> (c->arg ());
	deltap new_record = New refcounted<patch_delta> (a->file, a->patch);
	_logw->append (new_record);
	reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_TAG:
      {
	cfsw_tag_arg *a = static_cast<cfsw_tag_arg *> (c->arg ());
	deltap new_record = New refcounted<tag_delta> (a->commit, a->tag);
	_logw->append (new_record);
	reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_WRITE:
      {
	cfsw_write_arg *a = static_cast<cfsw_write_arg *> (c->arg ());
	deltap new_record = New refcounted<write_delta>
	  (a->file, a->direct, a->indirect,
	   a->double_indirect, a->triple_indirect, a->size);
	_logw->append (new_record);
	reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_CREATE:
    case CFSWPROC_MKDIR:
      {
	cfsw_create_arg *a = static_cast<cfsw_create_arg *> (c->arg ());
	cfsw_ftype type = (proc == CFSWPROC_CREATE) ? CFSW_REG : CFSW_DIR;
	ptr<inode_delta> new_inode;
	
	if (a->object != 0)
	  new_inode = New refcounted<inode_delta>
	    (type, a->object, a->attributes);
	else
	  new_inode = New refcounted<inode_delta>
	    (type, a->attributes);

	ptr<link_delta> new_link = New refcounted<link_delta>
	  (new_inode->object (), type, a->dir, a->name);

	cfsw_create_res *res = static_cast<cfsw_create_res *> (c->res ());
	res->object = new_inode->object ();
	assert (_logw);
	_logw->append (new_inode);
	_logw->append (new_link);
	reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_SYMLINK:
      {
	cfsw_symlink_arg *a = static_cast<cfsw_symlink_arg *> (c->arg ());

	if (a->linkto.len () > SYMLINK_MAX_LENGTH) {
	  reply (c, CFSRWERR_PATHTOOLONG);
	  return;
	}

	cfsw_ftype type = CFSW_LNK;
	ptr<inode_delta> new_inode;

	if (a->object != 0)
	  new_inode = New refcounted<inode_delta>
	    (type, a->object, a->attributes);
	else
	  new_inode = New refcounted<inode_delta>
	    (type, a->attributes);

	ptr<link_delta> new_link = New refcounted<link_delta> 
	  (new_inode->object (), type, a->dir, a->name);
	ptr<symlink_delta> new_linkto = New refcounted<symlink_delta>
	  (new_inode->object (), a->linkto);

	cfsw_symlink_res *res = static_cast<cfsw_symlink_res *> (c->res ());
	res->object = new_inode->object ();
	assert (_logw);
	_logw->append (new_inode);
	_logw->append (new_link);
	_logw->append (new_linkto);
	reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_LINK:
      {
	cfsw_link_arg *a = static_cast<cfsw_link_arg *> (c->arg ());
	ptr<link_delta> new_link =
	  New refcounted<link_delta> 
	  (a->object, CFSW_REG, a->dir, a->name);
	_logw->append (new_link);
        reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_RENAME:
      {
	cfsw_rename_arg *a = static_cast<cfsw_rename_arg *> (c->arg ());
	deltap new_record;
	if (a->target_exists)
	  new_record = New refcounted<rename_delta>
	    (a->object, a->type, a->from_dir, a->from_name,
	     a->to_dir, a->to_name, a->target);
	else
	  new_record = New refcounted<rename_delta>
	    (a->object, a->type, a->from_dir, a->from_name,
	     a->to_dir, a->to_name);
	_logw->append (new_record);
        reply (c, CFSRW_OK);
      }
      break;

    case CFSWPROC_RLOCK:
      {
	cfsw_rlock_arg *a = static_cast<cfsw_rlock_arg *> (c->arg ());
	const vervec &p = _replica->pkv ();
	if (p.size () == 0) {
	  if (!_anonymous)
	    _logw->rlock (a->handle, wrap (this, &cfsrw_impl::lock_cb, c));
	  else
	    _projinf->fetch_heads
	      (wrap (this, &cfsrw_impl::anon_lock_cb, c));
	}
	else {
	  if (!_anonymous)
	    _logw->rlock (a->handle, _replica->commits (), _replica->pkv (),
	                  wrap (this, &cfsrw_impl::lock_cb, c));
	  else
	    _projinf->fetch_heads 
	      (wrap (this, &cfsrw_impl::anon_lock_cb, c));
	}
      }
      break;

    case CFSWPROC_WLOCK:
      {
	cfsw_wlock_arg *a = static_cast<cfsw_wlock_arg *> (c->arg ());
	const vervec &p = _replica->pkv ();
	if (p.size () == 0)
	  _logw->prepare (a->handle, wrap (this, &cfsrw_impl::lock_cb, c));
	else
	  _logw->prepare (a->handle, _replica->commits (), _replica->pkv (),
	                  wrap (this, &cfsrw_impl::lock_cb, c));
      }
      break;

    case CFSWPROC_COMMIT:
      {
	assert (_replica->my_branch () >= 0);
	cfsw_commit_arg *a = static_cast<cfsw_commit_arg *> (c->arg ());
	if (!a->btag)
	  a->btag = "";

	btree_node x, f;
	// check if supersede point exists
        if ((a->flags & commit_record::MERGE_WITH) &&
	    !_replica->btree ()->lookup_by_commit (a->supersede, x)) {
	  assert (0);
	}
	// check if fork point exists
	if ((a->flags & commit_record::BRANCH_OFF) &&
	    !_replica->btree ()->lookup_by_commit (a->fork, f)) {
	  assert (0);
	}

	str pname = "";
	for (unsigned i=0; i<_projinf->nlogs (); i++) {
          if (_projinf->get_log_ptr (i) == _head) {
	    pname = _projinf->get_log_label (i) << ":" << _logw->seqn ();
	    break;
	  }
	}

	if (a->flags & commit_record::BRANCH_OFF) {
	  ptr<commit_record> cr = New refcounted<commit_record>
	    (f.commit, pname, a->message, a->btag, a->flags, a->supersede, a->commit_time);
	  _logw->commit (cr, wrap (this, &cfsrw_impl::commit_cb, c));
	}
	else {
	  ptr<commit_record> cr = New refcounted<commit_record>
	    (_replica->branch_head_info (_replica->my_branch ()),
	     pname, a->message, a->btag, a->flags, a->supersede, a->commit_time);
	  _logw->commit (cr, wrap (this, &cfsrw_impl::commit_cb, c));
	}
      }
      break;

    case CFSWPROC_ABORT:
      {
	_logw->abort (wrap (this, &cfsrw_impl::abort_cb, c, CFSRW_OK));
      }
      break;

    default:
      assert (0);
      break;
  }
}

void
cfsrw_impl::dispatch (cfsrcallp c)
{
  _replica->dispatch (c);
}

void
cfsr_memop (cfsr_proc p, cfsr_call_pair &c, bool free)
{
  switch (p) {
    case CFSRPROC_NULL:
      c.arg = 0;
      c.res = 0;
      break;
    case CFSRPROC_LOOKUP:
      if (free) {
	delete static_cast<cfsr_lookup_arg *> (c.arg);
	delete static_cast<cfsr_lookup_res *> (c.res);
      }
      else {
	c.arg = New cfsr_lookup_arg;
        c.res = New cfsr_lookup_res;
      }
      break;
    case CFSRPROC_GETATTR:
      if (free) {
	delete static_cast<cfsr_getattr_arg *> (c.arg);
	delete static_cast<cfsr_getattr_res *> (c.res);
      }
      else {
	c.arg = New cfsr_getattr_arg;
        c.res = New cfsr_getattr_res;
      }
      break;
    case CFSRPROC_FINDVER:
      if (free) {
	delete static_cast<cfsr_findver_arg *> (c.arg);
	delete static_cast<cfsr_findver_res *> (c.res);
      }
      else {
	c.arg = New cfsr_findver_arg;
        c.res = New cfsr_findver_res;
      }
      break;
    case CFSRPROC_READ:
      if (free) {
	delete static_cast<cfsr_read_arg *> (c.arg);
	delete static_cast<cfsr_read_res *> (c.res);
      }
      else {
	c.arg = New cfsr_read_arg;
        c.res = New cfsr_read_res;
      }
      break;
    case CFSRPROC_READLINK:
      if (free) {
	delete static_cast<cfsr_readlink_arg *> (c.arg);
	delete static_cast<cfsr_readlink_res *> (c.res);
      }
      else {
	c.arg = New cfsr_readlink_arg;
        c.res = New cfsr_readlink_res;
      }
      break;
    case CFSRPROC_READDIR:
      if (free) {
	delete static_cast<cfsr_readdir_arg *> (c.arg);
	delete static_cast<cfsr_readdir_res *> (c.res);
      }
      else {
	c.arg = New cfsr_readdir_arg;
        c.res = New cfsr_readdir_res;
      }
      break;
    default:
      assert (0);
      break;
  }
}

void
cfsw_memop (cfsw_proc p, cfsw_call_pair &c, bool free)
{
  switch (p) {
    case CFSWPROC_NULL:
      c.arg = 0;
      c.res = 0;
      break;
    case CFSWPROC_SETATTR:
      if (free) {
	delete static_cast<cfsw_setattr_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_setattr_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_UNLINK:
      if (free) {
	delete static_cast<cfsw_unlink_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_unlink_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_TAG:
      if (free) {
	delete static_cast<cfsw_tag_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_tag_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_PATCH:
      if (free) {
	delete static_cast<cfsw_patch_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_patch_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_WRITE:
      if (free) {
	delete static_cast<cfsw_write_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_write_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_CREATE:
    case CFSWPROC_MKDIR:
      if (free) {
	delete static_cast<cfsw_create_arg *> (c.arg);
	delete static_cast<cfsw_create_res *> (c.res);
      }
      else {
	c.arg = New cfsw_create_arg;
        c.res = New cfsw_create_res;
      }
      break;
    case CFSWPROC_SYMLINK:
      if (free) {
	delete static_cast<cfsw_symlink_arg *> (c.arg);
	delete static_cast<cfsw_symlink_res *> (c.res);
      }
      else {
	c.arg = New cfsw_symlink_arg;
        c.res = New cfsw_symlink_res;
      }
      break;
    case CFSWPROC_LINK:
      if (free) {
	delete static_cast<cfsw_link_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_link_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_RENAME:
      if (free) {
	delete static_cast<cfsw_rename_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_rename_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_RLOCK:
      if (free) {
	delete static_cast<cfsw_rlock_arg *> (c.arg);
	delete static_cast<cfsw_rlock_res *> (c.res);
      }
      else {
	c.arg = New cfsw_rlock_arg;
        c.res = New cfsw_rlock_res;
      }
      break;
    case CFSWPROC_WLOCK:
      if (free) {
	delete static_cast<cfsw_wlock_arg *> (c.arg);
	delete static_cast<cfsw_wlock_res *> (c.res);
      }
      else {
	c.arg = New cfsw_wlock_arg;
        c.res = New cfsw_wlock_res;
      }
      break;
    case CFSWPROC_ABORT:
      if (free) {
	delete static_cast<cfsw_abort_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_abort_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    case CFSWPROC_COMMIT:
      if (free) {
	delete static_cast<cfsw_commit_arg *> (c.arg);
	delete static_cast<cfsw_nores_res *> (c.res);
      }
      else {
	c.arg = New cfsw_commit_arg;
        c.res = New cfsw_nores_res;
      }
      break;
    default:
      assert (0);
      break;
  }
}

