/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "gcc_compat.h"
#include "stdinc.h"
#include "cfsrw.h"
#include "cfsr.h"
#include "cfs.h"
#include "cfsw_record.h"

extern int cfs_debug;

class cfsrw_write_obj
{
  str path;
  bigint fh;
  bigint va;
  cfsw_write_arg a;
  callback<void, cfsw_write_arg, cfsrw_status>::ref cb;
  bool local;
 
  int fd;
  unsigned wctr;
  unsigned werr;
  uint64 size;

  vec<bigint> direct;
  bigint indirect;
  bigint double_indirect;
  bigint triple_indirect;

  void done ()
  {
    if (wctr == 0) {
      close (fd);
      if (werr)
        cb (a, CFSRWERR_IO);
      else
        cb (a, CFSRW_OK);
      delete this;
    }
  }

  void
  insert_cb (ns_stat status, bigint key)
  {
    wctr --;
    if (status != NS_OK) {
      warn << path << ": dhash insert failed.\n";
      werr = 1;
    }
    done ();
  }

  bigint
  insert (char *data, size_t len)
  {
    bigint n;
    wctr ++;
    if (local) {
      n = nsobj->insert_local (data, len);
      delaycb (0, wrap (this, &cfsrw_write_obj::insert_cb, NS_OK, n));
    }
    else
      n = nsobj->insert_block
        (data, len, wrap (this, &cfsrw_write_obj::insert_cb));
    xfree (data);
    return n;
  }

  void do_write ()
  {
    direct.clear ();

    unsigned n = 0;
    bool fini = false;
    vec<bigint> fhs;

    while (!fini) {
      char buf [SN_BLKSIZE];
      int sz;

      if ((sz = read (fd, buf, SN_BLKSIZE)) < 0) {
        warn << "cannot read file " << path << "\n";
	werr = 1;
	done ();
        return;
      }

      n++;
      if (sz == 0)
        fini = true;
      
      if (sz > 0) {
        bigint h;
        write_block (buf, sz, &h);
        fhs.push_back (h);
	size += sz;
      }
    }
  
    unsigned blocknum = 0;
    while (blocknum < CFS_NDIR && blocknum < fhs.size()) {
      direct.push_back (fhs [blocknum]);
      blocknum++;
    }

    if (blocknum >= CFS_NDIR && blocknum < CFS_NDIR+SN_NFH)
      store_sind (&fhs, &indirect,
	          blocknum-CFS_NDIR, blocknum);
    if (blocknum >= CFS_NDIR+SN_NFH && blocknum < CFS_NDIR+SN_NFH+SN_NFH*SN_NFH)
      store_dind (&fhs, &double_indirect,
	          blocknum-CFS_NDIR-SN_NFH, blocknum);
    if (blocknum >= CFS_NDIR+SN_NFH+SN_NFH*SN_NFH)
      store_tind (&fhs, &triple_indirect,
	          blocknum-CFS_NDIR-SN_NFH-SN_NFH*SN_NFH, blocknum);

    a.file = fh;
    a.direct = direct;
    a.indirect = indirect;
    a.double_indirect = double_indirect;
    a.triple_indirect = triple_indirect;
    a.size = size;
    done ();
  }
 
  void 
  write_block (const char *block, ssize_t size, bigint *fh)
  {
    ptr<cfs_data> res = New refcounted<cfs_data> (CFS_FILEBLK);
    res->data->setsize (size);
    memcpy (res->data->base (), block, size);

    size_t xlen = 0;
    char *xbuf = NULL;
    xdrsuio x (XDR_ENCODE);

    if (xdr_cfs_data (x.xdrp (), res)) {
      xlen = x.uio ()->resid ();
      xbuf = suio_flatten (x.uio ());
    }

    *fh = insert (xbuf, xlen);
  }

  void
  store_sind (vec<bigint> *fhs, bigint *si_fh,
              uint32 off, uint32& blocknum)
  {
    uint32 blk = off / SN_NFH;
    ptr<cfs_data> sindir = New refcounted<cfs_data> (CFS_INDIR);
    sindir->indir->handles.setsize (SN_NFH);
  
    while (blk < SN_NFH && blocknum < fhs->size()) {
      sindir->indir->handles[blk] = (*fhs)[blocknum];
      blk++;
      blocknum++;
    }

    if (blk > 0) {
      size_t xlen = 0;
      char *xbuf = NULL;
      xdrsuio x (XDR_ENCODE);
      if (xdr_cfs_data (x.xdrp(), sindir)) {
        xlen = x.uio ()->resid ();
        xbuf = suio_flatten (x.uio ());
      }
      *si_fh = insert (xbuf, xlen);
    }
  }

  void
  store_dind (vec<bigint> *fhs, bigint *di_fh,
              uint32 off, uint32& blocknum)
  {
    ptr<cfs_data> dindir = New refcounted<cfs_data> (CFS_INDIR);
    dindir->indir->handles.setsize (SN_NFH);
    uint32 sblk = off/SN_NFH;
    uint32 oldbn = blocknum;

    while (sblk < SN_NFH && blocknum < fhs->size()) {
      off = (blocknum-CFS_NDIR)%SN_NFH;
      store_sind
        (fhs, &dindir->indir->handles[sblk], off, blocknum);
      sblk++;
    }

    if (blocknum > oldbn) {
      size_t xlen = 0;
      char *xbuf = NULL;
      xdrsuio x (XDR_ENCODE);
      if (xdr_cfs_data (x.xdrp(), dindir)) {
        xlen = x.uio ()->resid ();
        xbuf = suio_flatten (x.uio ());
      }
      *di_fh = insert (xbuf, xlen);
    }
  }

  void
  store_tind (vec<bigint> *fhs, bigint *ti_fh,
              uint32 off, uint32& blocknum)
  {
    ptr<cfs_data> tindir = New refcounted<cfs_data> (CFS_INDIR);
    tindir->indir->handles.setsize (SN_NFH);
    uint32 dblk = off/(SN_NFH*SN_NFH);
    uint32 oldbn = blocknum;

    while (dblk < SN_NFH && blocknum < fhs->size()) {
      off = blocknum%(SN_NFH*SN_NFH);
      store_dind
        (fhs, &tindir->indir->handles[dblk], off, blocknum);
      dblk++;
    }

    if (blocknum > oldbn) {
      size_t xlen = 0;
      char *xbuf = NULL;
      xdrsuio x (XDR_ENCODE);
      if (xdr_cfs_data (x.xdrp(), tindir)) {
        xlen = x.uio ()->resid ();
        xbuf = suio_flatten (x.uio ());
      }
      *ti_fh = insert (xbuf, xlen);
    }
  }

public:
  cfsrw_write_obj (str path, bigint fh, bigint va,
                   callback<void, cfsw_write_arg, cfsrw_status>::ref cb,
		   bool local)
    : path (path), fh (fh), va (va), cb (cb), local (local)
  {
    size = 0;
    wctr = 0;
    werr = 0;

    direct.clear ();

    fd = open (path, O_RDONLY);
    if (fd < 0) {
      if (cfs_debug)
	warn << "cfsrw_write_obj could not open " << path << " " << strerror(errno) << "\n";
      cb (a, CFSRWERR_IO);
      return;
    }
    do_write ();
  }
};

void
cfsrw_write_file (str path, bigint fh, bigint va,
                  callback<void, cfsw_write_arg, cfsrw_status>::ref cb)
{
  vNew cfsrw_write_obj (path, fh, va, cb, false);
}

void
cfsrw_write_file_local (str path, bigint fh, bigint va,
                        callback<void, cfsw_write_arg, cfsrw_status>::ref cb)
{
  vNew cfsrw_write_obj (path, fh, va, cb, true);
}


typedef callback<void,bigint,cfsrw_status>::ref cbbigint;

static void
get_sind (cbbigint bcb, unsigned nblocks, unsigned nread,
          callback<void, unsigned>::ref cb, ptr<cfs_data> data)
{
  if (!data) {
    bcb ((bigint) 0, CFSRWERR_IO);
    return;
  }

  for (unsigned i=0; i<SN_NFH && nread < nblocks; i++) {
    bigint n = data->indir->handles [i];
    bcb (n, CFSRW_OK);
    nread++;
  }

  if (nread == nblocks)
    bcb ((bigint) 0, CFSRWERR_NOMORELOG);
  else
    cb (nread);
}

static void
next_dind (cbbigint bcb, unsigned idx, unsigned nblocks, ptr<cfs_data> data,
           callback<void, unsigned>::ref cb, unsigned nread)
{
  assert (nread < nblocks);

  if (idx < SN_NFH) {
    bigint n = data->indir->handles [idx];
    bcb (n, CFSRW_OK);
    ns_retrieve_cfs_readonly
      (n, wrap (get_sind, bcb, nblocks, nread,
		wrap (next_dind, bcb, idx+1, nblocks, data, cb)));
  }
  else
    cb (nread);
}

static void
get_dind (cbbigint bcb, unsigned nblocks, unsigned nread,
          callback<void, unsigned>::ref cb, ptr<cfs_data> data)
{
  if (!data) {
    bcb ((bigint) 0, CFSRWERR_IO);
    return;
  }

  bigint n = data->indir->handles [0];
  bcb (n, CFSRW_OK);
  ns_retrieve_cfs_readonly
    (n, wrap (get_sind, bcb, nblocks, nread,
	      wrap (next_dind, bcb, 1, nblocks, data, cb)));
}

static void
next_tind (cbbigint bcb, unsigned idx, unsigned nblocks,
           ptr<cfs_data> data, unsigned nread)
{
  assert (nread < nblocks);

  if (idx < SN_NFH) {
    bigint n = data->indir->handles [idx];
    bcb (n, CFSRW_OK);
    ns_retrieve_cfs_readonly
      (n, wrap (get_dind, bcb, nblocks, nread,
		wrap (next_tind, bcb, idx+1, nblocks, data)));
  }
  else
    assert (0);
}

static void
get_tind (cbbigint bcb, unsigned nblocks, unsigned nread, ptr<cfs_data> data)
{
  if (!data) {
    bcb ((bigint) 0, CFSRWERR_IO);
    return;
  }

  bigint n = data->indir->handles [0];
  bcb (n, CFSRW_OK);
  ns_retrieve_cfs_readonly
    (n, wrap (get_dind, bcb, nblocks, nread,
	      wrap (next_tind, bcb, 1, nblocks, data)));
}

static void
check_tind (cbbigint bcb, deltap r, unsigned nblocks, unsigned nread)
{
  assert (nread < nblocks);
  write_delta *w = r->TEMPLATE record<write_delta> ();

  bcb (w->triple_indirect (), CFSRW_OK);
  ns_retrieve_cfs_readonly
    (w->triple_indirect (), wrap (get_tind, bcb, nblocks, nread));
}

static void
check_dind (cbbigint bcb, deltap r, unsigned nblocks, unsigned nread)
{
  assert (nread < nblocks);
  write_delta *w = r->TEMPLATE record<write_delta> ();

  bcb (w->double_indirect (), CFSRW_OK);
  ns_retrieve_cfs_readonly
    (w->double_indirect (),
     wrap (get_dind, bcb, nblocks, nread,
           wrap (check_tind, bcb, r, nblocks)));
}

void
iterate_write (deltap r, cbbigint bcb)
{
  // Iterates all the delta blocks and calls bcb on them ?
  write_delta *w = r->TEMPLATE record<write_delta> ();
  unsigned nblocks = w->size ()/SN_BLKSIZE;
  if ((w->size () % SN_BLKSIZE) != 0)
    nblocks++;
  unsigned nread = 0;

  for (unsigned i=0; i < w->ndirects () && i < nblocks; i++) {
    bcb (w->direct (i), CFSRW_OK);
    nread++;
  }

  if (w->ndirects () < nblocks) {
    bcb (w->indirect (), CFSRW_OK);
    ns_retrieve_cfs_readonly
      (w->indirect (),
       wrap (get_sind, bcb, nblocks, nread,
	     wrap (check_dind, bcb, r, nblocks)));
  }
  else
    bcb ((bigint) 0, CFSRWERR_NOMORELOG);
}

