/*
 *
 * Copyright (C) 2000 Frans Kaashoek (kaashoek@lcs.mit.edu)
 * Copyright (C) 2001 Frans Kaashoek (kaashoek@lcs.mit.edu) and 
 *                    Frank Dabek (fdabek@lcs.mit.edu).
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "chord.h"
#include "gcc_compat.h"
#include <misc_utils.h>
#include <id_utils.h>
#include <location.h>
#include <locationtable.h>
#include "comm.h"
#include "route.h"
#include <transport_prot.h>

#include <modlogger.h>
#define trace modlogger ("chord")

#include <configurator.h>

static struct chord_config_init {
  chord_config_init ();
} cci;

chord_config_init::chord_config_init ()
{
  bool ok = true;

#define set_int Configurator::only ().set_int
#define set_str Configurator::only ().set_str
  ok = ok && set_int ("chord.max_vnodes", 1024);

  ok = ok && set_int ("chord.nsucc", 16);
  ok = ok && set_int ("chord.npred", 1);
  ok = ok && set_int ("chord.ncoords", 3);

  ok = ok && set_str ("chord.rpc_mode", "stp");

  /** use the greedy metric instead.  Probably desirable if toes are
   *  enabled. */
  ok = ok && set_int ("chord.greedy_lookup", 0);
  /** try to terminate retrieves early if the hop has
   *  returned a "sufficient" number of successors.  */
  ok = ok && set_int ("chord.find_succlist_shaving", 1);

  ok = ok && set_int ("chord.checkdead_interval", 60);
  ok = ok && set_int ("chord.checkdead_max", 960);

  assert (ok);
#undef set_int
#undef set_str
}

chord::chord (str _wellknownhost, int _wellknownport, 
	      str _myname, int port, int max_cache,
	      ptr<restarter> r, str hostlistfile) :
  myname (_myname),
  nrcv (NULL),
  rpcm (NULL),
  _restarter (r),
  active (NULL)
{
  str rpcstr;
  bool ok;

  ok = Configurator::only ().get_int ("chord.max_vnodes", max_vnodes);
  assert (ok);
  
  ok = Configurator::only ().get_str ("chord.rpc_mode", rpcstr);
  assert (ok);

  chord_rpc_style = CHORD_RPC_STP;
  if (rpcstr == "tcp" || rpcstr == "TCP")
    chord_rpc_style = CHORD_RPC_SFST;
  else if (rpcstr == "udp" || rpcstr == "UDP")
    chord_rpc_style = CHORD_RPC_SFSU;

  str aqua_pk, hostlist;
  assert(Configurator::only().get_str("aqua.aqua_pk", aqua_pk));
  if (!hostlist::only().read_verifierpk(aqua_pk))
    fatal << "Could not read hostlist verifier from: " << aqua_pk << "\n";

  myport = startchord (port);
  warnx << "chord: running on " << myname << ":" << myport << "\n";

  nrcv = New refcounted<u_int32_t>;
  *nrcv = 0;
  if (chord_rpc_style == CHORD_RPC_SFSU)
    rpcm = New refcounted<rpc_manager> (nrcv);
  else if ((chord_rpc_style == CHORD_RPC_SFST) ||
           (chord_rpc_style == CHORD_RPC_SFSBT))
    rpcm = New refcounted<tcp_manager> (nrcv);
  else if (chord_rpc_style == CHORD_RPC_STP)
    rpcm = New refcounted<stp_manager> (nrcv);
  else
    fatal << "bad chord_rpc_style value: " << chord_rpc_style << "\n";

  locations = New refcounted<locationtable> (max_cache, rpcm);

  // add the wellknown host to the location table
  if (hostlistfile != "") {
    if (!hostlist::only().read_hostlist(hostlistfile))
      fatal << "Could not read hostlist from: " << hostlistfile << "\n";
    add_hosts_from_list(hostlistfile, myname, myport);
  }
  else if (myname != _wellknownhost || myport != _wellknownport) {
    chord_node node =
      prepare_node (_wellknownhost, _wellknownport, myname, myport);
    if (!(wellknown_node = locations->insert(node)))
      fatal << "Well known host failed to verify! Bailing.\n";
  }

  chordID me = hostlist::only().get_chordID(myname, myport);
  warn << "My ID is: " << me << "\n";
  hostlist::only().set_myid (me);
  locations->set_myid(me);

  nvnode = 0;
  srandom ((unsigned int) (getusec() & 0xFFFFFFFF));
}

void
chord::add_hosts_from_list(str hostlistfile, str hostname, int32_t port)
{
  warn << "hostlist has " << hostlist::only().size() << " entries\n";
  if (hostlist::only().size() < 1)
    fatal << "Hostlist is empty.\n";

  bool foundme = false;
  chord_node node;
  ptr<location> loc;
  struct hostentry *p;

  p = hostlist::only().first();
  while(p) {

    // Dont add my own location (this is added later in newvnode)
    if (p->name == hostname && p->port == port) {
      foundme = true;
      p = hostlist::only().next(p);
      continue;
    }

    node = prepare_node(p->name, p->port, hostname, port);
    warn << "adding to location table " << p->name << ":" << p->port << "\n";
    if (!(loc = locations->insert(node)))
      warn << "invalid host: " << p->name << ":" << p->port << "\n";

    p = hostlist::only().next(p);
  }

  if (!foundme) {
    fatal << "Could not find my entry in the hostlist: " 
	  << hostname << ":" << port << "\n";
  }
}

chord_node
chord::prepare_node(str _newname, int32_t _newport, 
		    str _myname, int32_t _myport)
{
  chord_node newnode;
  newnode.r.hostname = _newname;
  newnode.r.port = _newport ? _newport : _myport;
  newnode.x = hostlist::only().get_chordID(newnode.r.hostname, newnode.r.port);
  newnode.vnode_num = 0;
  newnode.coords.setsize (NCOORDS);
  // Make up some random initial information for this other node.
  for (int i = 0; i < NCOORDS; i++)
    newnode.coords[i] = (int) uniform_random_f (100000.0);

  return newnode;
}

void
chord::tcpclient_cb (int srvfd)
{
  int fd = accept (srvfd, NULL, NULL);
  if (fd < 0)
    warn << "chord: accept failed " << strerror (errno) << "\n";
  else {
    ptr<axprt> x = axprt_stream::alloc (fd, 1024*1025);

    ptr<asrv> s = asrv::alloc (x, transport_program_1);
    s->setcb (wrap (mkref(this), &chord::dispatch, s));
  }
}

int
chord::startchord (int myp, int type)
{
  
  int srvfd = inetsocket (type, myp);
  if (srvfd < 0)
    fatal ("binding %s port %d: %m\n",
	   (type == SOCK_DGRAM ? "UDP" : "TCP"), myp);
  
  _restarter->add_port(srvfd);
  if (myp == 0) {
    struct sockaddr_in addr;
    socklen_t len = sizeof (addr);
    bzero (&addr, sizeof (addr));
    if (getsockname (srvfd, (sockaddr *) &addr, &len) < 0) 
      fatal ("getsockname failed %m\n");
    myp = ntohs (addr.sin_port);
  }

  if (type == SOCK_DGRAM) {
    x_dgram = axprt_dgram::alloc (srvfd, sizeof(sockaddr), 230000);
    ptr<asrv> s = asrv::alloc (x_dgram, transport_program_1);
    s->setcb (wrap (mkref(this), &chord::dispatch, s));
  }
  else {
    int ret = listen (srvfd, 1000);
    assert (ret == 0);
    fdcb (srvfd, selread, wrap (this, &chord::tcpclient_cb, srvfd));
  }
  
  return myp;
}


int
chord::startchord (int myp)
{
  if ((chord_rpc_style == CHORD_RPC_SFST) ||
      (chord_rpc_style == CHORD_RPC_SFSBT))
  {
    // Ensure the DGRAM and STREAM sockets are on same port #,
    // since it is included in the Chord ID's hash.
    myp = startchord (myp, SOCK_STREAM);
  }

  return startchord (myp, SOCK_DGRAM);
}



ptr<vnode>
chord::newvnode (vnode_producer_t p, cbjoin_t cb)
{
  if (nvnode > max_vnodes)
    fatal << "Maximum number of vnodes (" << max_vnodes << ") reached.\n";
    
  chordID newID = hostlist::only().get_chordID (myname, myport, nvnode);

  warnx << gettime () << ": creating new vnode: " << newID << "\n";

  vec<float> coords;
  warn << gettime () << " coords are: ";
  for (int i = 0; i < NCOORDS; i++) {
    coords.push_back (uniform_random_f (1000000.0));
    warnx << (int) coords[i] << " " ;
  }
  warnx << "\n";
  ptr<location> l = locations->insert (newID, myname, myport, nvnode, coords);
  assert (l);
  if (wellknown_node == NULL)
    wellknown_node = l;
  locations->pin (newID);

  ptr<vnode> vnodep = (*p) (mkref (this), rpcm, l);
  
  if (!active) active = vnodep;
  nvnode++;
  vnodes.insert (newID, vnodep);
  
  vnodep->stabilize();
  (*cb) (vnodep, CHORD_OK);

  return vnodep;
}

void
chord::stats_cb (const chordID &k, ptr<vnode> v) { 
  v->stats();
}

void
chord::stats ()
{
  warnx << "CHORD NODE STATS\n";
  warnx << "# vnodes: " << nvnode << "\n";
  vnodes.traverse (wrap (this, &chord::stats_cb));
  rpcm->stats ();
}

void
chord::print_cb (strbuf outbuf, const chordID &k, ptr<vnode> v) {
  v->print (outbuf);
}

void
chord::print (strbuf &outbuf) {
  vnodes.traverse (wrap (this, &chord::print_cb, outbuf));
}

void
chord::stop_cb (const chordID &k, ptr<vnode> v) {
  v->stop ();
}

void
chord::stop () {
  vnodes.traverse (wrap (this, &chord::stop_cb));
}

const rpc_program *
chord::get_program (int progno)
{
  for (unsigned int i = 0; i < handledProgs.size (); i++)
    if (progno == (int)handledProgs[i]->progno)
      return handledProgs[i];
  return NULL;
}


bool
chord::isHandled (int progno) {
  for (u_int i = 0; i < handledProgs.size (); i++)
    if (progno == (int)handledProgs[i]->progno) return true;
  return false;
}
void
chord::handleProgram (const rpc_program &prog) {
  warn << "chord::handleProgram: " << prog.progno << "\n";
  if (isHandled (prog.progno)) return;
  else {
    handledProgs.push_back (&prog);
  }
}


void
chord::dispatch (ptr<asrv> s, svccb *sbp)
{
  if (!sbp) {
    s->setcb (NULL);
    return;
  }
  (*nrcv)++;
  
  dorpc_arg *arg = sbp->TEMPLATE getarg<dorpc_arg> ();

  switch (sbp->proc ()) {
  case TRANSPORTPROC_NULL:
    sbp->reply (NULL);
    break;
  case TRANSPORTPROC_DORPC:
    {
      //v (the destination chordID) is at the top of the header
      chordID *v = sbp->TEMPLATE getarg<chordID> ();
      vnode *vnodep = vnodes[*v];
      if (!vnodep) {
	trace << "unknown vnode " << *v << " for procedure "
	      << sbp->proc ()
	      << " (" << arg->progno << "." << arg->procno << ").\n";
	sbp->replyref (rpcstat (DORPC_UNKNOWNNODE));
	return;
      }
      
      //find the program
      const rpc_program *prog = get_program (arg->progno);
      if (!prog) {
	sbp->replyref (rpcstat (DORPC_NOHANDLER));
	return;
      }
      
      //unmarshall the args
      char *arg_base = (char *)(arg->args.base ());
      int arg_len = arg->args.size ();
      
      xdrmem x (arg_base, arg_len, XDR_DECODE);
      xdrproc_t proc = prog->tbl[arg->procno].xdr_arg;
      assert (proc);
      
      void *unmarshalled_args = prog->tbl[arg->procno].alloc_arg ();
      if (!proc (x.xdrp (), unmarshalled_args)) {
	warn << "dispatch: error unmarshalling arguments: "
	     << arg->progno << "." << arg->procno 
	     << " from " << *v <<"\n";
        xdr_delete (prog->tbl[arg->procno].xdr_arg, unmarshalled_args);
	sbp->replyref (rpcstat (DORPC_MARSHALLERR));
	return;
      }

      //call the handler
      user_args *ua = New user_args (sbp, unmarshalled_args, 
				     prog, arg->procno, arg->send_time);
      vnodep->fill_user_args (ua);
      if (!vnodep->progHandled (arg->progno)) {
	trace << "dispatch to vnode " << *v << " doesn't handle "
	      << arg->progno << "." << arg->procno << "\n";
	ua->replyref (chordstat (CHORD_NOHANDLER));
      } else {	      
	cbdispatch_t dispatch = vnodep->getHandler(arg->progno);
	(dispatch)(ua);
      }  
      
    }
    break;
  default:
    fatal << "Transport procedure " << sbp->proc () << " not handled\n";
  }
		
}



