/*
 *
 * Copyright (C) 2001  Frank Dabek (fdabek@lcs.mit.edu), 
 *                     Frans Kaashoek (kaashoek@lcs.mit.edu),
 *   		       Massachusetts Institute of Technology
 * 
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <sys/time.h>
#include <math.h>
#include <stdlib.h>

#include <sfsmisc.h>
#include <arpc.h>
#include <crypt.h>

#include <chord_types.h>
#include <chord.h>
#include <route.h>
#include <configurator.h>
#include <location.h>
#include <locationtable.h>
#include <merkle_misc.h>

#include "dhash_common.h"
#include "dhash.h"
#include "dhashcli.h"
#include "verify.h"
#include "download.h"
#include "dhash_store.h"
#include "succopt.h"

#include <coord.h>
#include <misc_utils.h>
#include <modlogger.h>

#define warning modlogger ("dhashcli", modlogger::WARNING)
#define info  modlogger ("dhashcli", modlogger::INFO)
#define trace modlogger ("dhashcli", modlogger::TRACE)

#ifdef DMALLOC
#include <dmalloc.h>
#endif
#include <ida.h>

static struct dhashcli_config_init {
  dhashcli_config_init ();
} dcci;

dhashcli_config_init::dhashcli_config_init ()
{
  bool ok = true;

#define set_int Configurator::only ().set_int

  /** Whether or not to order successors by expected latency */
  ok = ok && set_int ("dhashcli.order_successors", 1);

  assert (ok);
#undef set_int
}


// ---------------------------------------------------------------------------
// DHASHCLI

static u_int64_t tmark = getusec ();
 
dhashcli::dhashcli (ptr<vnode> node, str dhcs)
  : clntnode (node), ordersucc_ (true)
{
  int ordersucc = 1;
  Configurator::only ().get_int ("dhashcli.order_successors", ordersucc);
  ordersucc_ = (ordersucc > 0);
  // warn << "will order successors " << ordersucc_ << "\n";
}

void
dhashcli::checkpks_cb (cb_cpks cb, vec<chord_node> ss,
                       rpc_vec<bigint, RPC_INFINITY> keys,
                       rpc_vec<unsigned, RPC_INFINITY> vers,
                       ptr<dhash_check_pks_res> res, clnt_stat err)
{
  // u_int64_t t = getusec () - tmark;
  // warn << "checkpks: reply, +" << (long) t << "\n";

  vec<bigint> nv;
  vec<long> lv;

  if (err || res->status != DHASH_OK) {
    // try next successor, if any
    if (ss.size () == 0) {
      if (err)
	cb (DHASH_ERR, nv, lv);
      else
	cb (res->status, nv, lv);
    }
    else {
      chord_node s = ss.pop_front ();
      ptr<s_dhash_check_pks_arg> arg = New refcounted<s_dhash_check_pks_arg>;
      arg->keys = keys;
      arg->versions = vers;
      ptr<dhash_check_pks_res> res = New refcounted<dhash_check_pks_res> ();

      clntnode->doRPC 
        (s, dhash_program_1, DHASHPROC_CHECK_PKS, arg, res,
         wrap (this, &dhashcli::checkpks_cb, cb, ss, keys, vers, res), NULL);
    }
    return;
  }

  for (unsigned i=0; i<res->resok->keys.size (); i++)
    nv.push_back (res->resok->keys [i]);
  for (unsigned i=0; i<res->resok->versions.size (); i++)
    lv.push_back (res->resok->versions [i]);
  cb (DHASH_OK, nv, lv);
}

void
dhashcli::checkpks_rt_cb (cb_cpks cb, route_iterator *ci, bigint projkey,
                          rpc_vec<bigint, RPC_INFINITY> keys,
                          rpc_vec<unsigned, RPC_INFINITY> vers,
			  bool done)
{
  // u_int64_t t = getusec () - tmark;
  // warn << "checkpks: hop, +" << (long) t << "\n";

  // benjie: stop iterating if we already found the successor of the
  // key, we don't need to go to that node first.
  if (!done && ci->last_node ()->id () < projkey) {
    ci->next_hop ();
    return;
  }

  vec<chord_node> ss = ci->successors ();
  delete ci;

  chord_node s = ss.pop_front ();
  ptr<s_dhash_check_pks_arg> arg = New refcounted<s_dhash_check_pks_arg>;
  arg->projkey = projkey;
  arg->keys = keys;
  arg->versions = vers;
  ptr<dhash_check_pks_res> res = New refcounted<dhash_check_pks_res> ();

  clntnode->doRPC
    (s, dhash_program_1, DHASHPROC_CHECK_PKS, arg, res,
     wrap (this, &dhashcli::checkpks_cb, cb, ss, keys, vers, res), NULL);
}

void
dhashcli::checkpks (cb_cpks cb, bigint projkey,
                    rpc_vec<bigint, RPC_INFINITY> keys,
		    rpc_vec<unsigned, RPC_INFINITY> vers)
{
  // u_int64_t t = getusec () - tmark;
  // warn << "checkpks: +" << (long) t << "\n";

  route_iterator *ci = clntnode->produce_iterator_ptr (projkey, projkey);
  ci->first_hop
    (wrap (this, &dhashcli::checkpks_rt_cb, cb, ci, projkey, keys, vers), 0L);
}

void
dhashcli::retrieve_pk
  (blockID blockID, long vers, cb_ret cb, chordID projkey, int options)
{
  chordID myID = clntnode->my_ID ();
  trace << myID << ": retrieve (" << blockID << "): new retrieve.\n";
  ptr<rcv_state> rs = New refcounted<rcv_state> (blockID);
  rs->retpk = true;
  rs->vers = vers;
  rs->callbacks.push_back (cb);

  // u_int64_t t = getusec () - tmark;
  // warn << "retrieve_pk: +" << (long) t << "\n";

  rs->projkey = projkey;
  route_iterator *ci = clntnode->produce_iterator_ptr (projkey, projkey);
  ci->first_hop
    (wrap (this, &dhashcli::retrieve_block_hop_cb, rs, ci, options), 0L);
}

void
dhashcli::retrieve (blockID blockID, cb_ret cb, chordID projkey,
                    int options)
{
  chordID myID = clntnode->my_ID ();

  trace << myID << ": retrieve (" << blockID << "): new retrieve.\n";
  
  ptr<rcv_state> rs = New refcounted<rcv_state> (blockID);
  rs->callbacks.push_back (cb);
  rs->projkey = projkey;
  
  if (blockID.ctype == DHASH_KEYHASH) {
    route_iterator *ci = clntnode->produce_iterator_ptr (projkey, projkey);
    ci->first_hop
      (wrap (this, &dhashcli::retrieve_block_hop_cb, rs, ci, options), 0L);
  }
  else {
    vec<ptr<location> > sl = clntnode->succs ();
    clntnode->find_succlist
      (blockID.ID, projkey,
       // benjie: below line is left-over from chord, no longer used
       dhash::num_dfrags ()+(dhash::num_dfrags ()+1)/2,
       wrap (this, &dhashcli::retrieve_lookup_cb, rs), 0L);
  }
}

void
dhashcli::retrieve_block_hop_cb (ptr<rcv_state> rs, route_iterator *ci,
				 int options, bool done)
{
  // u_int64_t t = getusec () - tmark;
  // warn << "retrieve: block hop, +" << (long) t << "\n";

  // benjie: stop iterating if we already found the successor of the
  // key, we don't need to go to that node first.
  if (!done && ci->last_node ()->id () < rs->projkey) {
    ci->next_hop ();
    return;
  }

  chordID myID = clntnode->my_ID ();

  rs->timemark ();
  rs->r = ci->path ();
  rs->succs = ci->successors ();
  rs->errors = 0;
  rs->incoming_rpcs = 0;
  rs->blk = 0;
  dhash_stat status = ci->status () ? DHASH_CHORDERR : DHASH_OK;
  delete ci;

  if (status != DHASH_OK || rs->succs.size () < 1) {
    trace << myID << ": retrieve (" << rs->key
          << "): lookup failure: " << status << "\n";
    rs->complete (status, NULL); // failure
    rs = NULL;    
    return;
  }
  
  while (rs->succs.size () > dhash::pk_maxreps ())
    rs->succs.pop_back ();
  vec<float> lt;
  for (unsigned i=0; i<rs->succs.size (); i++) {
    ptr<location> l = clntnode->locations->lookup (rs->succs [i].x);
    if (l) lt.push_back (l->distance ());
    else lt.push_back (100000000);
  }
  vec<chord_node> succs;
  order_succs_by_latency (lt, rs->succs, succs);
  rs->succs = succs;
  rs->samev = false;

  unsigned n = ((int)floor(((double)dhash::pk_maxreps ())/2))+1;
  for (unsigned i=0; i<n && rs->succs.size ()>0; i++) {
    rs->incoming_rpcs++;
    chord_node s = rs->succs.pop_front ();
    if (rs->retpk)
      dhash_download::execute
        (clntnode, s, rs->key, rs->vers,
         wrap (this, &dhashcli::retrieve_dl_cb, rs));
    else
      dhash_download::execute
        (clntnode, s, rs->key, NULL, 0, 0, 0,
         wrap (this, &dhashcli::retrieve_dl_cb, rs));
  }
}

void
dhashcli::retrieve_dl_cb
  (ptr<rcv_state> rs, dhash_stat status, ptr<dhash_block> blk)
{
  // u_int64_t t = getusec () - tmark;
  // warn << "retrieve_pk: dl cb, +" << (long) t << "\n";

  chordID myID = clntnode->my_ID ();
  rs->incoming_rpcs --;

  if (blk && verify (rs->key.ID, rs->key.ctype, blk->data, blk->len)) {
    if (rs->blk == 0 || keyhash_version (rs->blk) < keyhash_version (blk))
      rs->blk = blk;
    rs->errors = 0;
  }
  else if (status == DHASH_SAMEVERSION) {
    if (rs->blk && keyhash_version (rs->blk) < rs->vers)
      rs->blk = 0;
    rs->samev = true;
    rs->errors = 0;
  }
  else if (status == DHASH_STALE) {
    if (rs->blk && keyhash_version (rs->blk) < rs->vers)
      rs->blk = 0;
    rs->stale = true;
    rs->errors = 0;
  }
  else if (rs->blk == 0 && rs->samev == false)
    rs->errors++;

  if (rs->incoming_rpcs == 0) {
    if (rs->blk) {
      rs->timemark ();
      rs->blk->ID = rs->key.ID;
      rs->blk->hops = rs->r.size ();
      rs->blk->errors = 0;
      rs->blk->retries = 0;
      rs->complete (DHASH_OK, rs->blk);
    }
    else if (rs->retpk && rs->samev)
      rs->complete (DHASH_SAMEVERSION, NULL);
    else if (rs->retpk && rs->stale)
      rs->complete (DHASH_STALE, NULL);
    else
      rs->complete (DHASH_NOENT, NULL);
    rs = NULL;
  }
}

void
dhashcli::fetch_frag (ptr<rcv_state> rs)
{
  register size_t i = rs->nextsucc;

  chordID myID = clntnode->my_ID ();

  // Ugh. No more successors available.
  if (i >= rs->succs.size ()) {
    // If there are outstanding fragments, there is still hope.
    // Hmm... Actually, this is a lie. If we know that they will not
    // provide us with enough total fragments to reconstruct the
    // block, e.g. incoming_rpcs + frags.size < NUM_DFRAGS, we should
    // just give up to the user now.  However for book keeping
    // purposes, we don't do this.
    if (rs->incoming_rpcs > 0)
      return;
    
    // Should we try harder? Like, try and get more successors and
    // check out the swath? No, let's just fail and have the higher
    // level know that they should retry.
    warning << myID << ": retrieve (" << rs->key 
            << "): out of successors; failing.\n";
    rs->complete (DHASH_NOENT, NULL);
    rs = NULL;
    return;
  }
  
  rs->incoming_rpcs += 1;

  dhash_download::execute (clntnode, rs->succs[i], 
			   blockID(rs->key.ID, rs->key.ctype, DHASH_FRAG),
			   (char *)NULL, 0, 0, 0, 
			   wrap (this, &dhashcli::retrieve_fetch_cb, rs, i),
			   wrap (this, &dhashcli::on_timeout, rs));
  rs->nextsucc += 1;
}

void
dhashcli::on_timeout (ptr<rcv_state> rs, 
		      chord_node dest,
		      int retry_num) 
{
#if 0
  if (retry_num == 1)
    fetch_frag (rs);
#endif
}

struct orderer {
  float d_;
  size_t i_;
  static int cmp (const void *a_, const void *b_) {
    const orderer *a = (orderer *) a_, *b = (orderer *) b_;
    return (int) (a->d_ - b->d_);
  }
};

#if 0
static void
order_succs (const vec<float> &me, const vec<chord_node> &succs,
	     vec<chord_node> &out)
{
  orderer *d2me = New orderer [succs.size()];
  vec<float> cursucc;
  for (size_t i = 0; i < succs.size (); i++) {
    cursucc.setsize (succs[i].coords.size ());
    for (size_t j = 0; j < succs[i].coords.size (); j++) 
      cursucc[j] = (float) succs[i].coords[j];
    d2me[i].d_ = Coord::distance_f (me, cursucc);
    d2me[i].i_ = i;
  }
  qsort (d2me, succs.size (), sizeof (*d2me), &orderer::cmp);
  out.clear ();
  for (size_t i = 0; i < succs.size (); i++) {
    out.push_back (succs[d2me[i].i_]);
  }
  delete[] d2me;
}
#endif

void
dhashcli::retrieve_lookup_cb (ptr<rcv_state> rs, vec<chord_node> succs,
			      route r, chordstat status)
{
  chordID myID = clntnode->my_ID ();
  rs->timemark ();
  rs->r = r;
  
  if (status) {
    trace << myID << ": retrieve (" << rs->key 
          << "): lookup failure: " << status << "\n";
    rs->complete (DHASH_CHORDERR, NULL); // failure
    rs = NULL;
    return;
  }
  
  while (succs.size () > dhash::ch_maxreps ())
    succs.pop_back ();

  if (succs.size () < dhash::num_dfrags ()) {
    warning << myID << ": retrieve (" << rs->key << "): "
	    << "insufficient number of successors returned!\n";
    rs->complete (DHASH_CHORDERR, NULL); // failure
    rs = NULL;
    return;
  }

  vec<float> lt;
  for (unsigned i=0; i<succs.size (); i++) {
    ptr<location> l = clntnode->locations->lookup (succs [i].x);
    if (l) lt.push_back (l->distance ());
    else lt.push_back (100000000);
  }
  order_succs_by_latency (lt, succs, rs->succs);

  // Dispatch NUM_DFRAGS parallel requests, even though we don't know
  // how many fragments will truly be needed. for small blocks, we
  // won't need NUM_DFRAGS, so extra fetch will allow us to react to
  // churn.
  u_int tofetch = dhash::num_dfrags ();
  if (tofetch > rs->succs.size ()) tofetch = rs->succs.size ();
  for (u_int i = 0; i < tofetch; i++)
    fetch_frag (rs);
}

void
dhashcli::retrieve_fetch_cb (ptr<rcv_state> rs, u_int i, dhash_stat stat,
			     ptr<dhash_block> block)
{
  chordID myID = clntnode->my_ID ();

  rs->incoming_rpcs -= 1;
  if (rs->completed) {
    // Here it might just be that we got a fragment back after we'd
    // already gotten enough to reconstruct the block.
    // warn << myID << ": EXTRA FRAGMENT (" << rs->key << ")\n";
    return;
  }

  if (!block) {
    warn << myID << ": retrieve failed (" 
	 << armor64A(rs->key.ID) << " " << rs->key.ID << ")\n";
    rs->errors++;
    fetch_frag (rs);
    return;
  }
  
  str frag (block->data, block->len);
  for (size_t j = 0; j < rs->frags.size (); j++) {
    if (rs->frags[j] == frag) {
      warning << myID << ": retrieve " 
	      << armor64A(rs->key.ID) << " " << rs->key.ID
	      << " duplicate fragment retrieved from successor " << i+1
	      << "; same as fragment " << j << "\n";
      rs->errors++;
      fetch_frag (rs);
      return;
    }
  }
  rs->frags.push_back (frag);
  rs->nfrags++;

  strbuf newblock;

  if (!Ida::reconstruct (rs->frags, newblock)) {
    if (rs->nfrags >= dhash::num_dfrags ()) {
      warning << myID << ": retrieve " << armor64A(rs->key.ID) << " " 
	      << rs->key.ID << " reconstruction failed.\n";
      rs->errors++;
      fetch_frag (rs);
    }
    return;
  }
  
  str tmp (newblock);
  if (!verify (rs->key.ID, rs->key.ctype, tmp.cstr (), tmp.len ())) {
    if (rs->nfrags >= dhash::num_dfrags ()) {
      warning << myID << ": retrieve " << armor64A(rs->key.ID)
	      << " " << rs->key.ID << " verify failed.\n";
      rs->errors++;
      fetch_frag (rs);
    }
    return;
  }
  
  rs->timemark ();
  ptr<dhash_block> blk = 
    New refcounted<dhash_block> (tmp.cstr (), tmp.len (), rs->key.ctype);
  blk->ID = rs->key.ID;
  blk->hops = rs->r.size ();
  blk->errors = rs->errors;
  blk->retries = blk->errors;
  
  for (size_t i = 1; i < rs->times.size (); i++) {
    timespec diff = rs->times[i] - rs->times[i - 1];
    blk->times.push_back (diff.tv_sec * 1000 +
			  int (diff.tv_nsec/1000000));
  }
  
  rs->complete (DHASH_OK, blk);
  rs = NULL;
}

void
dhashcli::insert (ref<dhash_block> block, cbinsert_path_t cb,
                  chordID projkey, int options)
{
  if (block->ctype == DHASH_CONTENTHASH) {
    vec<ptr<location> > sl = clntnode->succs ();
    lookup
      (block->ID, projkey,
       wrap (this, &dhashcli::insert_lookup_cb, block, cb, options, projkey));
  }
  else
    lookup
      (projkey, projkey,
       wrap (this, &dhashcli::insert_lookup_cb, block, cb, options, projkey));
}

void
dhashcli::insert_succlist_cb
  (ref<dhash_block> block, cbinsert_path_t cb,
   chordID projkey, int options, vec<chord_node> succs, chordstat status)
{
  if (status) {
    vec<chordID> rrr;
    cb (DHASH_CHORDERR, rrr);
    return;
  }

  route r;
  r.push_back (clntnode->locations->lookup (projkey));
  insert_lookup_cb (block, cb, options, projkey, DHASH_OK, succs, r);
}

void
dhashcli::insert_lookup_cb
  (ref<dhash_block> block, cbinsert_path_t cb, int options,
   chordID projkey, dhash_stat status, vec<chord_node> succs, route r)
{
  vec<chordID> mt;
  if (status) {
    (*cb) (DHASH_CHORDERR, mt);
    return;
  }
  
  unsigned replicas_needed = dhash::num_dfrags ();
  unsigned replicas_wanted = dhash::ch_maxreps ();

  if (block->ctype == DHASH_KEYHASH) {
    replicas_needed = ((int)floor(((double)dhash::pk_maxreps ())/2))+1;
    replicas_wanted = dhash::pk_maxreps ();
  }

  if (succs.size () < replicas_needed) {
    // this is a failure condition, since we can't hope to reconstruct
    // the block reliably.
    info << "Not enough successors for insert: |succs| " << succs.size ()
         << ", need " << replicas_needed << "\n";
    warn << "DHASH_STOREERR A\n";
    (*cb) (DHASH_STOREERR, mt);
    return;
  }

  if (succs.size () < replicas_wanted)
    // benjie: this is not a failure condition, since if we don't
    // receive ch_maxreps number of store replies in insert_store_cb,
    // we are still allowed to proceed.
    info << "Number of successors less than desired: |succs| "
         << succs.size () << ", wanted " << replicas_wanted << "\n";

  while (succs.size () > replicas_wanted)
    succs.pop_back ();

  ref<sto_state> ss = New refcounted<sto_state> (block, cb);
  ss->succs = succs;

  if (block->ctype == DHASH_KEYHASH) {
    for (u_int i = 0; i < succs.size (); i++) {
      // Count up for each RPC that will be dispatched
      ss->out += 1;
      
      ptr<location> dest = clntnode->locations->lookup_or_create (succs[i]);
      dhash_store::execute
	(clntnode, dest, blockID (block->ID, block->ctype, DHASH_BLOCK),
	 projkey, block,
	 wrap (this, &dhashcli::insert_store_cb,
	       ss, r, i, ss->succs.size (), replicas_needed),
	 i == 0 ? DHASH_STORE : DHASH_REPLICA);
    }
    return;
  }

  str blk (block->data, block->len);

  // Cap the maximum.
  u_long m = Ida::optimal_dfrag (block->len, dhash::dhash_mtu ());
  if (m > dhash::num_dfrags ())
    m = dhash::num_dfrags ();

  // trace << "Using m = " << m << " for block size " << block->len << "\n";

  for (u_int i = 0; i < succs.size (); i++) {
    str frag = Ida::gen_frag (m, blk);
    
    ref<dhash_block> blk = New refcounted<dhash_block> 
      ((char *)NULL, frag.len (), DHASH_CONTENTHASH);
    bcopy (frag.cstr (), blk->data, frag.len ());
    
    // Count up for each RPC that will be dispatched
    ss->out += 1;

    ptr<location> dest = clntnode->locations->lookup_or_create (succs[i]);

    dhash_store::execute (clntnode, dest,
			  blockID (block->ID, block->ctype, DHASH_FRAG),
			  projkey, blk,
			  wrap (this, &dhashcli::insert_store_cb,
				ss, r, i, succs.size (),
				dhash::num_dfrags ()),
			  DHASH_FRAGMENT);
  }
}

void
dhashcli::insert_store_cb (ref<sto_state> ss, route r, u_int i,
			   u_int nstores, u_int min_needed,
			   dhash_stat err, chordID id, bool present)
{
  ss->out -= 1;
  if (err) {
    info << "fragment/block store failed: " << ss->block->ID
	 << " fragment " << i << "/" << nstores
	 << ": " << err << "\n";
  } else {
    info << "fragment/block store ok: " << ss->block->ID
	 << " fragment " << i << "/" << nstores
	 << " at " << id << "\n";
    ss->good += 1;
  }

  // Count down until all outstanding RPCs have returned
  vec<chordID> r_ret;

  if (ss->out == 0) {
    chordID myID = clntnode->my_ID ();
    if (ss->good < nstores) {
      warning << myID << ": store (" << ss->block->ID << "): only stored "
	      << ss->good << " of " << nstores << " encoded.\n";
      if (ss->good < min_needed) {
	warning << myID << ": store (" << ss->block->ID << "): failed;"
	  " insufficient frags/blocks stored.\n";
	
	r_ret.push_back (ss->succs[0].x);
	warn << "DHASH_STOREERR B\n";
	(*ss->cb) (DHASH_STOREERR, r_ret);
	// We should do something here to try and store this fragment
	// somewhere else.
	return;
      }
    }
    
    for (unsigned int i = 0; i < r.size (); i++)
      r_ret.push_back (r[i]->id ());
    (*ss->cb) (DHASH_OK, r_ret);
  }
}

void
dhashcli::lookup (chordID n, bigint projkey, dhashcli_lookupcb_t cb)
{
#if 0
  clntnode->find_successor
    (blockID, wrap (this, &dhashcli::lookup_findsucc_cb, blockID, cb));
#else
  route_iterator *ci = clntnode->produce_iterator_ptr (n, projkey);
  ci->first_hop
    (wrap (this, &dhashcli::lookup_hop_cb, n, cb, ci), 0L);
#endif
}

void
dhashcli::lookup_hop_cb (chordID n, dhashcli_lookupcb_t cb, 
                         route_iterator *ci, bool done)
{
  // benjie: stop iterating if we already found the successor of the
  // key, we don't need to go to that node first.
  if (!done && ci->last_node ()->id () < n) {
    ci->next_hop ();
    return;
  }
  
  vec<chord_node> s = ci->successors ();
  route r = ci->path ();
  delete ci;
  (*cb) (DHASH_OK, s, r);
}

void
dhashcli::lookup_findsucc_cb (chordID blockID, dhashcli_lookupcb_t cb,
			      vec<chord_node> s, route path, chordstat err)
{
  if (err) 
    (*cb) (DHASH_CHORDERR, s, path);
  else
    (*cb) (DHASH_OK, s, path);
}


void
dhashcli::sendblock (ptr<location> dst, blockID bid_to_send,
                     chordID projkey, ptr<dbrec> blk,
		     callback<void, dhash_stat, bool>::ref cb)
{
  ref<dhash_block> dhblk = New refcounted<dhash_block> 
    (blk->value, blk->len, bid_to_send.ctype);

  dhash_store::execute 
    (clntnode, dst, bid_to_send, projkey, dhblk,
     wrap (this, &dhashcli::sendblock_cb, cb),
     bid_to_send.dbtype == DHASH_BLOCK ? DHASH_REPLICA : DHASH_FRAGMENT);
}

void
dhashcli::sendblock_cb (callback<void, dhash_stat, bool>::ref cb, 
			  dhash_stat err, chordID dest, bool present)
{
  (*cb) (err, present);
}

