/*
 *
 * Copyright (C) 2001  Frank Dabek (fdabek@lcs.mit.edu), 
 *                     Frans Kaashoek (kaashoek@lcs.mit.edu),
 *   		       Massachusetts Institute of Technology
 * 
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "gcc_compat.h"
#include <chord.h>
#include <chord_types.h>
#include <dhashgateway_prot.h>
#include <location.h>

#include "dhash_common.h"
#include "dhash.h"
#include "dhashcli.h"
#include "dhashgateway.h"
#include <configurator.h>

#include "arpc.h"
#ifdef DMALLOC
#include "dmalloc.h"
#endif


// ------------------------------------------------------------------------
// DHASHGATEWAY

dhashgateway::dhashgateway (ptr<axprt_stream> x, ptr<chord> node, bool overload):
  _overload(overload)
{
  clntsrv = asrv::alloc (x, dhashgateway_program_1,
	                 wrap (mkref (this), &dhashgateway::dispatch));
  dhcli = New refcounted<dhashcli> (node->active);
}

dhashgateway::~dhashgateway ()
{
  extern int current_clients;
  current_clients --;
}

void
dhashgateway::dispatch (svccb *sbp)
{
  if (!sbp) {
    // setting clntsrv to 0 removes the last reference to this gateway
    // object, stored in the asrv object's callback.
    clntsrv = 0;
    return;
  }

  switch (sbp->proc ()) {
  case DHASHPROC_NULL:
    sbp->reply (NULL);
    return;

  case DHASHPROC_INSERT:
    {
      dhash_insert_arg *arg = sbp->TEMPLATE getarg<dhash_insert_arg> ();

      if (_overload) {
	dhash_insert_res res(DHASH_BUSY);
	sbp->reply(&res);
      }

      ref<dhash_block> block =
	New refcounted<dhash_block> (arg->block.base (), arg->len, arg->ctype);
      block->ID = arg->blockID;
      dhcli->insert (block,
	             wrap (mkref (this), &dhashgateway::insert_cb, sbp),
		     arg->projkey, arg->options);
    }
    break;
    
  case DHASHPROC_RETRIEVE:
    {
      dhash_retrieve_arg *arg = sbp->TEMPLATE getarg<dhash_retrieve_arg> ();

      if (_overload) {
	dhash_retrieve_res res(DHASH_BUSY);
	sbp->reply(&res);
      }
        
      dhcli->retrieve
	(blockID (arg->blockID, arg->ctype, DHASH_BLOCK),
	 wrap (mkref (this), &dhashgateway::retrieve_cb, sbp),
	 arg->projkey, arg->options);
    }
    break;

  case DHASHPROC_RETRIEVEPK:
    {
      dhashgw_retrieve_pk_arg *arg =
	sbp->TEMPLATE getarg<dhashgw_retrieve_pk_arg> ();

      if (_overload) {
	dhash_retrieve_res res(DHASH_BUSY);
	sbp->reply(&res);
      }

      dhcli->retrieve_pk
	(blockID (arg->key, DHASH_KEYHASH, DHASH_BLOCK), (long) arg->vers,
	 wrap (mkref (this), &dhashgateway::retrieve_cb, sbp),
	 arg->projkey, arg->options);
    }
    break;

  case DHASHPROC_CHECKPKS:
    {
      if (_overload) {
	dhash_check_pks_res res(DHASH_BUSY);
	sbp->reply(&res);
      }

      dhashgw_check_pks_arg *arg =
	sbp->TEMPLATE getarg<dhashgw_check_pks_arg> ();
      dhcli->checkpks
	(wrap (mkref (this), &dhashgateway::checkpks_cb, sbp),
	 arg->projkey, arg->keys, arg->versions);
    }
    break;

  case DHASHPROC_SEND_STATS:
    {
      dhashgw_send_stats_arg *arg =
	sbp->TEMPLATE getarg<dhashgw_send_stats_arg> ();
      str new_entries (arg->stats.base(), arg->len);

      int err = append_stats (new_entries);
      dhashgw_send_stats_res res (DHASH_OK);
      if (err)
	res.set_status (DHASH_ERR);

      sbp->reply (&res);
    }
    break;

  default:
    sbp->reject (PROC_UNAVAIL);
    break;
  }
}

void
dhashgateway::checkpks_cb (svccb *sbp, dhash_stat status,
                           vec<bigint> keys, vec<long> vers)
{
  dhashgw_check_pks_res res (status);
  if (!status) {
    res.resok->keys.setsize (keys.size ());
    for (unsigned i=0; i<keys.size (); i++)
      res.resok->keys [i] = keys [i];
    res.resok->versions.setsize (vers.size ());
    for (unsigned i=0; i<vers.size (); i++)
      res.resok->versions [i] = (unsigned) vers [i];
  }
  sbp->reply (&res);
}

void
dhashgateway::insert_cb (svccb *sbp, dhash_stat status, vec<chordID> path)
{
  dhash_insert_res res (status);
  if (status == DHASH_OK) {
    res.resok->path.setsize (path.size ());
    for (unsigned int i = 0; i < path.size (); i++)
      res.resok->path[i] = path[i];
  }
  sbp->reply (&res);
}

#define SET_RETRIEVE_REPLY \
  if (!block) \
    res.set_status (stat); \
  else { \
    res.resok->block.setsize (block->len); \
    res.resok->ctype = block->ctype; \
    res.resok->len = block->len; \
    res.resok->hops = block->hops; \
    res.resok->errors = block->errors; \
    res.resok->retries = block->retries; \
    res.resok->path.setsize (path.size ()); \
    for (u_int i = 0; i < path.size (); i++) \
      res.resok->path[i] = path[i]->id (); \
    res.resok->times.setsize (block->times.size ()); \
    for (u_int i = 0; i < block->times.size (); i++) \
      res.resok->times[i] = block->times[i]; \
    memcpy (res.resok->block.base (), block->data, block->len); \
  }

void
dhashgateway::retrieve_cb (svccb *sbp, dhash_stat stat,
                           ptr<dhash_block> block, route path)
{
  dhash_retrieve_res res (DHASH_OK);
  SET_RETRIEVE_REPLY;
  sbp->reply (&res);
}

int 
dhashgateway::append_stats (str new_stats)
{
  str stat_file;
  if (!Configurator::only().get_str ("dhash.stats_file", stat_file)) 
    stat_file = "stats.log";

  int fd = open (stat_file.cstr(), O_WRONLY | O_APPEND | O_CREAT | O_NONBLOCK, 0644);
  if (fd < 0) {
    warn << "error opening stat file: " << stat_file << "\n";
    return -1;
  }
  
  size_t sent = 0;
  while (sent < new_stats.len()) {
    int r = write (fd, new_stats.cstr() + sent, new_stats.len()-sent);
    if (r < 0) {
      warn << "Could not record stats, error " << errno 
	   << " while appending to stat log: " << stat_file << "\n";
      return -1;
    }
    sent += r;
  }
  close (fd);
  return 0;
}
