/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include "pastwatch.h"
#include "opts.h"
#include "pdb.h"
#include "path.h"
#include "io.h"

typedef callback<void, bigint, int>::ref fhcb_t;
extern int check_dirent (str dirpath, dirent *e, bool& ignore_entry);
extern void checkout_one (bigint, str, str, bigint, cbi);
extern void commit (bigint root, str path, cbi cb);
extern void checkin (const pde_t& e, fhcb_t fhcb);

extern vec<str> notify_cmds;
extern vec<str> notify_files;

static bool
dir_empty (str path)
{
  DIR *dir = opendir (path);
  if (!dir)
    return false;
  struct dirent *e;
  while ((e = readdir (dir))) {
    if (strcmp (e->d_name, ".") && strcmp (e->d_name, "..") &&
        strcmp (e->d_name, PASTWATCH_REF)) {
      closedir (dir);
      return false;
    }
  }
  closedir (dir);
  return true;
}

static void 
add_dir_reply (str path, str cfspath, cbi cb, bigint fh, int status)
{
  if (status) {
    warn << "cannot add " << path << " to repository\n";
    cb (1);
    return;
  }

  warn << "Marking " << path << " to be added\n";
  pmgr->schedule_add (path, pfe (fh));

  // for commit notification message
  notify_cmds.push_back ("A");
  notify_files.push_back (path);
  
  ptr<pfe> x = pmgr->lookup (get_parent_dir (path));
  if (x) {
    x->cr_addr = 0;
    pmgr->schedule_add (get_parent_dir (path), *x);
  }

  pmgr->schedule_mkdir (path, cfspath);
  cb (0);
}

void
add (bigint root, str path, cbi cb)
{
  bool removed = false;
  bigint fh = 0;
  bigint cr_addr;
  char w = 0;

  struct dirops dpv;
  str name;
  str dir = get_parent_dir (path, name);

  ptr<pfe> p = pmgr->lookup (dir);
  if (!p) {
    warn << "directory " << dir
         << " is not in the repository\n";
    cb (0);
    return;
  }

  if (get_dirops (dir, dpv) == 0) {
    for (unsigned i=0; i< dpv.remove.size () && !removed; i++) {
      if (dpv.remove [i]->name == name) {
	removed = true;
	fh = dpv.remove [i]->fh;
	cr_addr = dpv.remove [i]->cr_addr;
	w = 'R';
      }
    }
  }

  struct stat sb;
  int err = lstat (path, &sb);
    
  if (err) {
    if (removed) {
      warn << "resurrecting " << path << "\n";
      del_dirop (path, w);
      str cfspath;
      get_cfspath (path, cfspath);
      checkout_one (fh, path, cfspath, cr_addr, cb);
    }
    else {
      warn << "cannot stat(2) " << path << "\n";
      cb (0);
    }
    return;
  }
  else if (removed) {
    pastwatch_status
      ("C", path, "should be removed and is still here");
    cb (0);
    return;
  }

  // check if directory already been added
  if (sb.st_mode & S_IFDIR) {
    str d = path << "/" << PASTWATCH_REF;
    if (lstat (d, &sb) == 0) {
      warn << path << " contains " << PASTWATCH_REF << ", already added?\n";
      cb (0);
      return;
    }
  }

  // check if file already exists in repository or not
  if (pmgr->exists (path)) {
    warn << path << " in database, already exists in repository?\n";
    cb (0);
    return;
  }
 
  if (sb.st_mode & S_IFDIR) {
    // a directory, actually add it to repository
    arg_readonly = false;
    warn << "adding " << path << " to repository\n";

    str dcp;
    get_cfspath (dir, dcp);
    str cfspath = dcp << "/" << name;

    pde_t e;
    e.name  = name;
    e.type  = CFSDIR;
    e.path  = path;
    e.dirfh = p->fh;
    e.fh    = 0;
    checkin (e, wrap (add_dir_reply, path, cfspath, cb));
  }
  else {
    ptr<add_dirop> a = New refcounted<add_dirop> (name);
    new_dirop (dir, a);
    cb (0);
  }
}

static int remove_work = 0;
static int error = 0;

static void
done (cbi cb)
{
  if (remove_work == 0)
    cb (error);
}

static void
expand (str p, vec<str>& x)
{
  DIR *dir = opendir (p);
  if (dir) {
    struct dirent *e;
    while ((e = readdir (dir))) {
      bool should_ignore = false;
      if (check_dirent (p, e, should_ignore))
	continue;
      if (!should_ignore) {
	str fp;
	if (p == ".")
	  fp = e->d_name;
	else
	  fp = p << "/" << e->d_name;
	if (e->d_type == DT_DIR) {
	  if (arg_recurse)
	    expand (fp, x);
	}
	else
	  x.push_back (fp);
      }
    }
    closedir (dir);
  }
}

static void 
rmdir_reply (str path, cbi cb, cfswcallp c, cfsrw_status status)
{
  remove_work--;
  if (status) {
    warn << "cannot remove " << path << " from repository\n";
    error = -1;
    done (cb);
    return;
  }

  // for commit notification message
  notify_cmds.push_back ("R");
  notify_files.push_back (path);
    
  pmgr->schedule_remove (path);
  if (pmgr->remove_dir (path))
    warn << "cannot remove " << path << ", perhaps it is not empty?\n";
  done (cb);
}

static void
remove_lookup_reply (str path, cbi cb, cfsrcallp cc, cfsrw_status status)
{
  remove_work--;
  if (status) {
    warn << "cannot lookup directory " << path << " in repository\n";
    error = -1;
    done (cb);
    return;
  }

  cfsr_lookup_arg *arg = static_cast<cfsr_lookup_arg *> (cc->arg ());
  cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (cc->res ());

  remove_work++;
  cfswcallp c = New refcounted<cfswcall>
    (CFSWPROC_UNLINK, wrap (rmdir_reply, path, cb));
  cfsw_unlink_arg *a = static_cast<cfsw_unlink_arg *> (c->arg ());
  a->object = res->object;
  a->type = CFSW_DIR;
  a->dir = arg->dir;
  a->name = arg->name;
  cfsw_call (c);
}

static void
readdir_reply (bigint root, str path,
               cbi cb, cfsrcallp c, cfsrw_status status)
{
  remove_work--;
  if (status) {
    warn << "cannot read directory " << path << " in repository\n";
    error = -1;
    done (cb);
    return;
  }
  cfsr_readdir_res *res = static_cast<cfsr_readdir_res *> (c->res ());
  if (res->dir->entries != 0L) {
    warn << path << " not empty in repository.  Pastwatch will only remove empty directories from the repository\n";
    done (cb);
    return;
  }

  str name;
  str dir = get_parent_dir (path, name);
  
  // an empty directory, actually remove it to repository
  arg_readonly = false;
  warn << "removing " << path << " from repository\n";

  str dcp;
  get_cfspath (dir, dcp);
  str cfspath = dcp << "/" << name;
      
  remove_work++;
  cfsr_lookup
    (root, cfspath, wrap (remove_lookup_reply, path, cb));
}

static void
do_remove (bigint root, str path, bigint fh, bigint cr_addr, cbi cb)
{
  remove_work--;
  struct stat sb;
  int err = lstat (path, &sb);

  // check if the file or directory to be removed is still around; if
  // -f is given, remove the file or directory
  if (!err) {
    if (arg_force) {
      if (sb.st_mode & S_IFDIR) {
	if (dir_empty (path)) {
	  // check if dir is empty in cfs
          remove_work++;
          cfsrcallp c = New refcounted<cfsrcall>
            (CFSRPROC_READDIR, wrap (readdir_reply, root, path, cb));
          cfsr_readdir_arg *arg = static_cast<cfsr_readdir_arg *> (c->arg ());
          arg->dir = fh;
          arg->cookie = 0;
          arg->count = 2048;
          cfsr_call (cr_addr, c);
	  return;
	}
	else {
	  warn << path << ": directory not empty, cannot remove.\n";
	  done (cb);
	  return;
	}
      }
      else if (unlink (path)) {
	perror ("unlink");
        warn << "cannot remove " << path << " from working directory.\n";
	done (cb);
	return;
      }
    }
    else {
      warn << "please remove " << path << " from working directory first.\n";
      done (cb);
      return;
    }
  }

  str name;
  str dir = get_parent_dir (path, name);
  ptr<remove_dirop> a =
    New refcounted<remove_dirop> (name, fh, cr_addr);
  new_dirop (dir, a);

  warn << "Marking " << path << " to be removed\n";
  pmgr->schedule_remove (path);
  done (cb);
}

void
remove (bigint root, str p, cbi cb)
{
  vec<str> files;
    
  struct stat sb;
  int err = lstat (p, &sb);
  if (!err && (sb.st_mode & S_IFDIR) && !arg_dironly)
    expand (p, files);
  else
    files.push_back (p);

  if (files.empty ())
    warn << "no files to remove\n";

  while (!files.empty ()) {
    str p = files.pop_front ();
    ptr<pfe> x = pmgr->lookup (p);
    if (!x) {
      warn << "cannot find " << p << " in database\n";
      continue;
    }
    remove_work++;
    delaycb (0, wrap (do_remove, root, p, x->fh, x->cr_addr, cb));
  }

  done (cb);
}

