/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include "pastwatch.h"
#include "opts.h"
#include "path.h"
#include "pdb.h"
#include "io.h"

extern ptr<cfsrw> _cfsrw;

extern int  check_dirent (str dirpath, dirent *e, bool& ignore_entry);
extern void read_file (int, bigint, bigint, callback<void, cfsrw_status>::ref);
extern int  diff_context (const char *f1, const char *f2, strbuf &b, str t1, str t2);
extern int  diff   (const char *f1, const char *f2, strbuf &b);
extern int  isdiff (const char *f1, const char *f2);
extern int  isconf (const char *path);
extern str  get_tag_of_commit (bigint cn);

#define ISDIR(sb) ((sb.st_mode & S_IFDIR) == S_IFDIR)

class diff_object : public virtual refcount {
private:
  vec<str> objects;
  str path;
  cbi cb;
  bigint fh;
  unsigned rvn[2];
  bigint rv[2];
  bigint rep_cr_addr[2];
  bigint tag_cr_addr[2];
  unsigned work;
  int fd[2];
  str fn[2];

  unsigned user_specified_versions () 
  {
    unsigned versions = 0;

    for(unsigned i=0; i<2; i++)
      if (arg_tag[i]      != "" || 
	  arg_revision[i] != 0  || 
	  arg_date[i]     != ""   )
	versions++;
    return versions;
  }

  void read_done (unsigned index, cfsrw_status status)
  {
    work--;

    close (fd[index]);
    if (status) {
      unlink (fn[index]);
      warn << "cannot read version " << arg_revision[0] << " of "
	   << path << " from repository\n";
      next ();
      return;
    }

    // Wait until we retrieved all the files.
    if (work != 0)
      return;

    if (isdiff (fn[0].cstr (), fn[1].cstr ()) > 0) {
      str d0 = get_tag_of_commit (rv[0]);
      str d1 = (user_specified_versions() > 1) ? get_tag_of_commit(rv[1]) : path;
      strbuf out;
      out << "File: " << path << "\n";
      out << "===============================================" << "\n";

      if (arg_contextdiff)
        out << "diff -c -t " << d0 << " " << d1 << "\n";
      else
        out << "diff -t " << d0 << " " << d1 << "\n";
      out.tosuio ()->output (1);

      str opts = "";
      strbuf b;
      if (arg_contextdiff) {
	strbuf f0, f1;
	str n;
        get_parent_dir (path, n);
	f0 << "r" << rvn[0] << " of " << n;
	f1 << "r" << rvn[1] << " of " << n;
	diff_context (fn[0].cstr (), fn[1].cstr (), b, f0, f1);
      }
      else
	diff (fn[0].cstr (), fn[1].cstr (), b);
      b.tosuio ()->output (1);
    }

    unlink (fn[0]);
    if (user_specified_versions() == 2)
      unlink (fn[1]); 

    next ();
  }

  void
  getattr (unsigned index, cfsrcallp c, cfsrw_status status)
  {
    if (status) {
      warn << "cannot get attribute of " << path << "\n";
      next ();
      work--;
      return;
    }
    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (c->res ());
    rvn[index] = res->attr->revision;
    rv[index] = res->attr->commit;

    fn[index] = pastwatch_tmpfile ();
    fd[index] = open (fn[index], O_CREAT | O_TRUNC | O_WRONLY, 0600);
    read_file
      (fd[index], fh, rep_cr_addr[index], wrap (mkref (this), &diff_object::read_done, index));
  }

  void
  find_fh (unsigned index)
  {
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR, wrap (mkref (this), &diff_object::getattr, index));
    cfsr_getattr_arg *a = static_cast<cfsr_getattr_arg *> (c->arg ());
    a->object = fh;
    cfsr_call (rep_cr_addr[index], c);
  }

  void
  findver_cb (unsigned index, cfsrcallp c, cfsrw_status status)
  {
    if (status) {
      warn << "cannot find revision " << arg_revision[index]
	   << " of " << path << "\n";
      next ();
      work--;
      return;
    }
    cfsr_findver_res *res = static_cast<cfsr_findver_res *> (c->res ());
    rep_cr_addr[index] = res->cr_addr;
    find_fh (index);
  }

  void
  expand (str path)
  {
    DIR *dir = opendir (path);
    if (dir == 0)
      return;

    struct dirent *e;
    while ((e = readdir (dir))) {
      bool should_ignore = false;
      if (check_dirent (path, e, should_ignore))
	continue;
      if (!should_ignore) {
	str fp;
	if (path == ".")
	  fp = e->d_name;
	else
	  fp = path << "/" << e->d_name;
	if (e->d_type == DT_DIR && arg_recurse)
	  expand (fp);
	else
	  objects.push_back (fp);
      }
    }
    closedir (dir);
  }

  void 
  begin_fetch(unsigned index) 
  {
    rep_cr_addr[index] = tag_cr_addr[index]; 
    if (arg_revision[index]) {
      // If the revision was specified.
      cfsrcallp c = New refcounted<cfsrcall>
	(CFSRPROC_FINDVER, wrap (mkref (this), &diff_object::findver_cb, index));
      cfsr_findver_arg *a = static_cast<cfsr_findver_arg *> (c->arg ());
      a->object = fh;
      a->revision = arg_revision[index];
      cfsr_call (rep_cr_addr[index], c);
    } else
      find_fh (index);
  }

  void
  next () 
  {
    if (objects.size () == 0) {
      cb (0);
      return;
    }

    path = objects.pop_front ();
    ptr<pfe> v = pmgr->lookup (path);
    if (!v) {
      warn << "no information for " << path << ", not checked-in?\n";
      next ();
      return;
    }
    if (isconf (path) != 0) {
      warn << path << " contains conflicts\n";
      next ();
      return;
    }
    fh = v->fh;

    switch (user_specified_versions()) {
    case 0:
      work=1;
      fn[1] = path.cstr();

      if (v->sticky != 0)
	rep_cr_addr[0] = v->sticky;
      else
	rep_cr_addr[0] = my_branch_head (_cfsrw->getreplica ());
      find_fh (0);
      break;

    case 1:
      work=1;
      fn[1] = path.cstr();
      begin_fetch(0);
      break;

    case 2:
      work=2;
      begin_fetch(0);
      begin_fetch(1);
      break;
    default:
      break;
    }
  }

public:
  diff_object (str path, cbi cb)
    : path (path), cb (cb), work (0)
  {
    fn[0] = "";
    fn[1] = "";
    fd[0] = -1;
    fd[1] = -1;
    struct stat sb;
    if (lstat (path, &sb)) {
      warn << "cannot open " << path << "\n";
      cb (-1);
      return;
    }
    else if (ISDIR (sb))
      expand (path);
    else
      objects.push_back (path);

    // If the user specified any tags, figure out which snapshot 
    // those tags refer to.  Assumes that arg_tag was already translated
    // to a commit_record address.
    for(unsigned i=0; i<2; i++) {
      if (arg_tag[i] != "") {
	bigint n;
	dearmor64A_bigint (arg_tag[i], n);
	ptr<btree_srv> srv = _cfsrw->getreplica ()->btree ();
	btree_node x;
	srv->lookup_by_commit (n, x);
	tag_cr_addr[i] = x.commit.rptr;
      }
    }

    next ();
  }
};

void
diff (bigint root, str path, cbi cb)
{
  ptr<diff_object> a = New refcounted<diff_object> (path, cb);
}

