/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <errno.h>
#include <rxx.h>
#include "pastwatch.h"
#include "opts.h"
#include "path.h"
#include "adb.h"
#include "pdb.h"

#define DIROP_LINE_LEN 1100

static dirop_factory diropf;

bool
add_dirop::parse_entry (const char *s)
{
  vec<str> n;
  split (&n, "/", s+2);

  if (n.size () == 2) {
    name = n [0];
    if (!dearmor64A_bigint (n [1], fh))
      return false;
  }
  else
    name = n [0];
  return true;
}

bool
remove_dirop::parse_entry (const char *s)
{
  fh = 0;
  vec<str> n;
  split (&n, "/", s+2);
  if (n.size () == 3) {
    name = n [0];
    if (!dearmor64A_bigint (n [1], fh))
      return false;
    if (!dearmor64A_bigint (n [2], cr_addr))
      return false;
    return true;
  }
  return false;
}

int
new_dirop (str dir, ptr<dirop> x)
{
  str p = get_ops_file (dir);
  ptr<adb> db = New refcounted <adb> ();
  if (db->opendb (p, &diropf, true)) {
    warn << "cannot open " << p << "\n";
    return 1;
  }
  str k = x->getkey ();
  db->insert (k, x);
  return db->closedb ();
}

int
get_dirops (str dir, struct dirops &d)
{
  struct stat sb;
  if (lstat (dir, &sb) || (sb.st_mode & S_IFDIR) != S_IFDIR) {
    warn << dir << " is not a directory\n";
    return 1;
  }

  str p = get_ops_file (dir);
  ptr<adb> db = New refcounted <adb> ();
  if (db->opendb (p, &diropf, true)) {
    warn << "cannot open " << p << "\n";
    return 1;
  }

  for (unsigned i=0; i<db->entries.size (); i++) {
    if ((*(reinterpret_cast<ptr<dirop>*>(&db->entries [i])))->which == 'A')
      d.push_add (*reinterpret_cast<ptr<add_dirop>*> (&db->entries [i]));
    else if ((*(reinterpret_cast<ptr<dirop>*>(&db->entries [i])))->which == 'R')
      d.push_remove (*reinterpret_cast<ptr<remove_dirop>*> (&db->entries [i]));
    else
      assert (0);
  }

  return 0;
}

int
set_dirops (str dir, struct dirops &d)
{
  struct stat sb;
  if (lstat (dir, &sb) || (sb.st_mode & S_IFDIR) != S_IFDIR) {
    warn << dir << " is not a directory\n";
    return 1;
  }

  str p = get_ops_file (dir);
  ptr<adb> db = New refcounted <adb> ();
  if (db->opendb (p, &diropf, true)) {
    warn << "cannot open " << p << "\n";
    return 1;
  }
  db->modified = true;
  db->entries.clear ();
  for (unsigned i=0; i<d.add.size (); i++)
    db->insert (d.add [i]->getkey (), d.add [i]);
  for (unsigned i=0; i<d.remove.size (); i++)
    db->insert (d.remove [i]->getkey (), d.remove [i]);
  return db->closedb ();
}

static str
getkey (char which, str n)
{
  str x;
  if (which == 'A')
    x = "A " << n;
  else if (which == 'R')
    x = "R " << n;
  else
    assert (0);
  return x;
}

int
del_dirop (str p, char which)
{
  str n;
  str dir = get_parent_dir (p, n);
  str dbf = get_ops_file (dir);
  ptr<adb> db = New refcounted <adb> ();
  if (db->opendb (dbf, &diropf, true)) {
    warn << "cannot open " << dbf << "\n";
    return 1;
  }
  db->del (getkey (which, n));
  return db->closedb ();
}

int
del_dirops (const vec<str> &v, const vec<char> &w)
{
  for (unsigned i=0; i<v.size (); i++) {
    str p = v [i];
    str n;
    str dir = get_parent_dir (p, n);
    str dbf = get_ops_file (dir);
    ptr<adb> db = New refcounted <adb> ();
    if (db->opendb (dbf, &diropf, true)) {
      warn << "cannot open " << dbf << "\n";
      return 1;
    }
    db->del (getkey (w [i], n));
    db->closedb ();
  }
  return 0;
}

int
get_cfspath (str path, str &cfspath)
{
  struct stat sb;
  
  str refpath;
  str name;
  int r = lstat (path, &sb);

  if (r == 0 && (sb.st_mode & S_IFDIR)) {
    name = "";
    refpath = get_info_file (path);
  }
  else {
    str s = get_dir (path);
    refpath = get_info_file (s);

    if (s == ".")
      name = path;
    else if (s == "")
      name = "";
    else
      name = substr (path, s.len ()+1);
  }

  str buf = file2str (refpath);
  if (!buf)
    return -1;
  
  vec<str> n;
  split (&n, "\n", buf);
  assert (n.size () > 1);
  if (n [0] == "" || n [0] == "\n")
    cfspath = name;
  else {
    strbuf b;
    if (name == "")
      b << n [0];
    else
      b << n [0] << "/" << name;
    cfspath = b;
  }
  return 0;
}

