
/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/time.h>
#include "stdinc.h"
#include "parseopt.h"
#include "qhash.h"
#include "rxx.h"
#include <dhash_common.h>
#include "ns.h"
#include "vervec.h"
#include "traverse.h"
#include "projinterface.h"
#include "cfsw_prot.h"
#include "record.h"
#include "cfsw_record.h"
#include "snapshot.h"
#include "cfs.h"

extern void iterate_write
(deltap r, callback<void,bigint,cfsrw_status>::ref cb);

class syncer : public virtual refcount {
private:
  ptr<ns> _nsobj;
  ptr<projinterface> _projinf;
  bigint _projid;
  callback<void,int>::ref final_cb;
  int _work;
  vec<bigint> _newpks;
  vec<bigint> _newcommits;
  vec<bigint> _depends;
  vec<bigint> _stop;
  comvec _logheads;
  ptr<ns_block> _insert_this_commit;
  bigint _insert_this_commit_hash;
  bool _ispush;

  // If a crec addr is in here, that means we already enqueued it to
  // be checked.
  qhash<bigint, bool> _done; 

  vec<deltap> _deltas;
  vec<ptr<ns_block> > _delta_blocks;
  vec<bigint> _delta_hashes;
  unsigned _ndeltas;
  
  vec<str> _rlist;

  void done(int err)
  {
    if (err) {
      warn << "Encountered an error during sync\n";
      final_cb (-1);
    } else {
      warn << "sync done.\n";
      final_cb (0);
    }
  }
  
  void
  insert_pks_cb (ns_stat status, bigint n)
  {
    if (status) {
      warn << "cannot insert pk " << armor64A (n) << " into DHT\n";
      _nsobj = 0;
      // Encountered an error inserting a pk block.
      done(-1);
      return;
    }
    else
      write_pks ();
  }

  void
  write_pks ()
  {
    if (_newpks.size ()) {
      bigint n = _newpks.pop_front ();
      warn << "inserting pk " << armor64A (n) << "\n";
      _nsobj->insert_raw_pk_from_cache
	(n, wrap (mkref(this), &syncer::insert_pks_cb));
    }
    else {
      // No more pks to insert
      // I think this is the exit point when everything is complete
      // succesfully.
      _nsobj = 0;
      done(0);
      return;
    }
  }

  void
  stage2 (int err)
  {
    /* Stage 2 writes the pk blocks to the DHT.
     */
    if (_work == 0) {
      if (err == 0)
	write_pks ();
      else {
        // Encountered an error
	done(-1);
	return;
      }
    }
  }

  void
  insert_done (ns_stat status, bigint n)
  {
    _work--;
    if (status) {
      warn << "cannot insert " << armor64A (n) << "\n";
      stage2 (-1);
      return;
    }
    if (_work == 0)
      check_commits ();
  }

  void
  get_write_block (bigint n, ptr<ns_block> b)
  {
    _work--;
    if (!b) {
      warn << "cannot fetch write block!\n";
      stage2 (-1);
      return;
    }
    else {
      _work++;
      bigint x = _nsobj->insert_block
	(b->data, b->len, wrap (mkref(this), &syncer::insert_done), true);
      warn << "block inserted: " << x << "\n";
      assert (x == n);
    }
  }

  void
  dump_write_cb (bigint n, cfsrw_status status)
  {
    if (status) {
      if (status != CFSRWERR_NOMORELOG) {
        warn << "iterate write failed!\n";
	stage2 (-1);
	return;
      }

      // done inserting deltas
      assert (status == CFSRWERR_NOMORELOG);
      _work--;
      stage2 (0);

    } else {
      _work++;
      get_write_block (n, _nsobj->retrieve_local (n));
    }
  }
  
  void
  dump_write (deltap r)
  {
    _work++;
    iterate_write (r, wrap (mkref(this), &syncer::dump_write_cb));
  }
  
  void
  fetch_deltap_cb (bigint x, unsigned i, ptr<ns_block> b)
  {
    _work--;
    
    _ndeltas++;
    if (_ndeltas == _deltas.size ())
      _work--;

    if (!b) {
      warn << "cannot load record " << armor64A (x) << "\n";
      stage2 (-1);
      return;
    }
  
    str s (b->data, b->len);
    deltap r = cfsw_deltap_parse (s);
    if (!r) {
      warn << "cannot load record " << armor64A (x) << "\n";
      stage2 (-1);
      return;
    }

    _deltas [i] = r;
    _delta_blocks [i] = b;
  
    if (_ndeltas == _deltas.size ()) {
      for (unsigned i=0; i<_deltas.size (); i++) {
        _work++;
        bigint xx = _nsobj->insert_block
          (_delta_blocks [i]->data, _delta_blocks [i]->len,
	   wrap (mkref(this), &syncer::insert_done), true);
	warn << "inserted delta: " << armor64A(xx) << "\n";
	assert (xx == _delta_hashes [i]);
        if (_deltas [i]->rtype () == cfsw_delta::T_WRITE)
          dump_write (_deltas [i]);
      }
    }
  }

  void
  fetch_ind_cb (bigint n, ptr<ns_block> b)
  {
    _work--;
    if (!b) {
      warn << "cannot load indirect " << armor64A (n) << "\n";
      stage2 (-1);
      return;
    }
    _work++;
    _nsobj->insert_block
      (b->data, b->len, wrap (mkref(this), &syncer::insert_done), true);
    warn << "indirect inserted: " << n << "\n";
  }

  void
  get_ind_cb (crec, bigint n)
  {
    _work++;
    _nsobj->retrieve_block (n, wrap (mkref(this), &syncer::fetch_ind_cb, n));
  }
  
  void
  get_deltas_cb (crec c, bigint n, bool failed)
  {
    if (failed) {
      _work--;
      warn << "cannot iterate log in " << armor64A (n) << "\n";
      stage2 (-1);
      return;
    }

    unsigned i = _delta_hashes.size ();
    _delta_hashes.push_back (n);
    _work++;
    _nsobj->retrieve_block (n, wrap (mkref(this), &syncer::fetch_deltap_cb, n, i));
  }

  void
  get_cached_commit (bool inlog, bigint c, bool from_dht, ptr<ns_block> b)
  {
    // get_cached_commit is called for a commit record 
    // If the commti record exists in the DHT, the <from_dht> equals true.
    // else, the commit record is still loaded, but we get the cached
    // block and <from_dht> equal false.

    if (!b) { // yikes
      warn << "cannot fetch " << armor64A (c) << " from replica!\n";
      stage2 (-1);
      return;
    }

    crec r = commit_record::alloc (str (b->data, b->len));
    if (!r) { // yikes
      warn << "cannot parse " << armor64A (c) << " into commit record!\n";
      stage2 (-1);
      return;
    }

    // Figure out which log this commit record belongs to.
    bool haslog = false;
    unsigned logidx = 0;
    for (unsigned i=0; i<_logheads.size (); i++) {
      if (r->depends ().head == _logheads.head (i)) {
	logidx = i;
	commit_info ci = r->depends();
	warn << armor64A (ci.rptr) << " is in a known log\n";
	haslog = true;
	break;
      }
    }

    // If this crec is part of the log, and we haven't seen its
    // parent, then add its parent to be processed.
    if (inlog && r->prev () != 0) {
      assert (haslog);
      if (r->prev () != _stop [logidx]) {
        bool *x = _done [r->prev ()];
        if (!x) {
	  _newcommits.push_back (r->prev ());
	  _done.insert (r->prev (), true);
        }
      }
    }

    // I believe r->depends().rptr is the same as r->prev()

    if (r->depends ().rptr != 0) {
      if (!haslog) {
        bool *x = _done [r->depends ().rptr];
        if (!x) {
	  commit_info ci = r->depends();
	  warn << armor64A (ci.rptr) << " in unknown log\n";

	  if (!inlog && from_dht && !_ispush) {
	    // if this block itself is not in a known log, and it is
	    // in the DHT, and we are not in push mode, then we assume
	    // blocks it depends on are in the DHT as well, so dont 
	    // push this block's parent to the DHT.
	  }
	  else {
	    _depends.push_back (r->depends ().rptr);
	    _done.insert (r->depends ().rptr, true);
	  }
        }
      }
    }

    if (!from_dht || _ispush) {
      warn << "inserting " << armor64A (c) << "'s blocks\n";
      assert (_insert_this_commit == 0);
      _insert_this_commit = b;
      _insert_this_commit_hash = c;
      if (r->ndeltas () > 0) {
        _deltas.clear ();
        _delta_blocks.clear ();
        _delta_hashes.clear ();
        _ndeltas = 0;
        _deltas.setsize (r->ndeltas ());
        _delta_blocks.setsize (r->ndeltas ());
        _work++;
        get_deltas 
	  (r, wrap (mkref(this), &syncer::get_ind_cb, r),
	   wrap (mkref(this), &syncer::get_deltas_cb, r));
      }
    }

    if (_work == 0)
      check_commits ();
  }

  void
  check_commit (bool inlog, bigint c, ptr<ns_block> b)
  {
    if (b) {
    // Got the block from DHT, not from cache
      warn << armor64A (c) << ": in DHT\n";
      get_cached_commit (inlog, c, true, b);
    }
    else {
      warn << armor64A (c) << ": not in DHT\n";
      get_cached_commit (inlog, c, false, _nsobj->retrieve_local (c));
    }
  }

  void
  check_commits_next ()
  {
    assert (_work == 0);

    if (_newcommits.size () > 0) {
      bigint c = _newcommits.pop_front ();
      _nsobj->retrieve_block
	(c, wrap (mkref(this), &syncer::check_commit, true, c), true);
    }
    else if (_depends.size () > 0) {
      bigint c = _depends.pop_front ();
      _nsobj->retrieve_block
	(c, wrap (mkref(this), &syncer::check_commit, false, c), true);
    }
    else
      stage2 (0);
  }

  void
  insert_commit_done (ns_stat status, bigint n)
  {
    if (status) {
      warn << "cannot insert " << armor64A (n) << "\n";
      stage2 (-1);
      return;
    }
    check_commits_next ();
  }

  void
  check_commits ()
  {
    assert (_work == 0);

    if (_insert_this_commit) {
      ptr<ns_block> b = _insert_this_commit;
      _insert_this_commit = 0;
      bigint n = _nsobj->insert_block
        (b->data, b->len, wrap (mkref(this), &syncer::insert_commit_done), true);
      assert (n == _insert_this_commit_hash);
    }
    else
      check_commits_next ();
  }

  void
  check_log_ptr (unsigned i, ptr<head_state> v, ptr<keyhash_payload> p)
  {
    _work--;
    
    lptr u = 0;
    if (!p || !(u = log_pointer::alloc (p->buf ()))) {
      warn << armor64A (v->commits.head (i))
	   << ": lptr not available on DHT\n";
      _newpks.push_back (v->commits.head (i));
      _stop [i] = 0;
    }
    else {
      if (p->version () < _projinf->get_log_dhash_version(i)) {
        warn << "stale pk in dht " << armor64A (v->commits.head (i)) 
	     << " " << p->version() << " < " <<  _projinf->get_log_dhash_version(i) << "\n";
        _newpks.push_back (v->commits.head (i));
	_stop [i] = u->commit ();
      }
      else
	// log-head on DHT and replica the same, or replica has stale
	// version
	_stop [i] = v->commits.rptr (i);
    }

    if (_work == 0) {
      _logheads = v->commits;
      for (unsigned i=0; i<_logheads.size (); i++) {
	if (_logheads.rptr(i) != 0 && 
	    (_ispush || _logheads.rptr (i) != _stop [i])) {
	  _newcommits.push_back (_logheads.rptr (i));
	  _done.insert (_logheads.rptr (i), true);
	}
      }
      check_commits ();
    }
  }

  void
  got_version (ptr<head_state> v)
  { 
    // If the DHT version is newer, then we just got the updated version.
    // Here, we check if the DHT version is older than the local version because
    // _projinf->fetch would have gotten the newer, local version already.
    assert(_work == 0);
    _stop.setsize (v->commits.size ());

    // check log pointers
    for (unsigned i=0; i<v->commits.size (); i++) {
      _work++;
      _nsobj->retrieve_pk_from_net
	(v->commits.head (i), wrap (mkref(this), &syncer::check_log_ptr, i, v));
    }
  }

  void
  check_pkblocks (bigint block_id, long version, ptr<head_state> p, ptr<keyhash_payload> payload)
  {
    _work--;
    if (!payload || payload->version() < version)
      _newpks.push_back (block_id);
    if (_work == 0)
      got_version(p);
  }

  void
  got_projinf (bigint projid, ptr<head_state> p, log_status status)
  {    
    if (status) {
      warn << "Cannot load project block\n";
      stage2 (-1);
      return;
    }

    // check if the projblock and user blocks are available in DHT
    _work++;
    _nsobj->retrieve_pk_from_net 
      (projid, wrap (mkref(this), 
		     &syncer::check_pkblocks, projid, _projinf->get_projblk_version(), p));
    for(unsigned i=0; i<_projinf->nusers(); i++) {
      _work++;
      _nsobj->retrieve_pk_from_net 
	(_projinf->get_user_id(i), 
	 wrap (mkref(this), &syncer::check_pkblocks, _projinf->get_user_id(i), 
	       _projinf->get_userblk_version (i), p));
    }
  }

public:
  syncer (ptr<ns> nsobj, bigint projid, bool push, callback<void,int>::ref cb)
    : _nsobj(nsobj), _projid(projid), final_cb(cb), _ispush (push)
  {
    _work = 0;
    _ndeltas = 0;
    _insert_this_commit = 0;
    
    if (!nsobj->connected ()) {
      warn << "Cannot connect to Aqua, sync cannot run in disconnect mode.\n";
      final_cb (-1);
      return;
    }
    _projinf = New refcounted<projinterface>(projid);
    _projinf->fetch_heads(wrap(mkref(this), &syncer::got_projinf, projid));
  }

  ~syncer () 
  {
    warn << "syncer " << armor64A (_projid) << " cleaned up\n";
  }
};


/*
 * sync should be readonly.  It changes the repository and the replica,
 * but it does not create any new commit records.
 */

void
sync_repository (ptr<ns> nsobj, bigint projid, bool push_mode, callback<void, int>::ref cb)
{
  ptr<syncer> a = New refcounted<syncer> (nsobj, projid, push_mode, cb);
}
